# Copyright 2015-2018 Massimo Riello, Adam Fekete, Fawzi Mohamed, Ankit Kariryaa
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import logging
import os
import sys
import operator
import copy

# Python 2 compability
from io import open

from .LammpsCommon import get_metaInfo
from nomadcore.caching_backend import CachingLevel
from nomadcore.simple_parser import SimpleMatcher as SM
from nomadcore.simple_parser import mainFunction
import nomadcore.ActivateLogging

from re import escape as esc



class sformat(object):
    float = r"[-+]?\d*\.?[0-9]*([Ee][+-]?[0-9]+)?"
    int = r"[-+]?\d+"




############################################################
# This is the parser for the DATA file of LAMMPS.
############################################################

logger = logging.getLogger(name="nomad.LammpsDataParser")





class LammpsDataParserContext(object):

    """Context for parsing LAMMPS DATA file.

    Attributes:
        # dos_energies: Stores parsed energies.
        # dos_values: Stores parsed DOS values.

    The onClose_ functions allow processing and writing of cached values after a section is closed.
    They take the following arguments:
        backend: Class that takes care of writing and caching of metadata.
        gIndex: Index of the section that is closed.
        section: The cached values and sections that were found in the section that is closed.
    """

    def __init__(self, converter, writeMetaData=True):
        """Args:
            writeMetaData: Determines if metadata is written or stored in class attributes.
        """
        self.parser = None
        self.converter = converter
        self.writeMetaData = writeMetaData

    def startedParsing(self, fInName, parser):
        """Function is called when the parsing starts and the compiled parser is obtained.

        Args:
            fInName: The file name on which the current parser is running.
            parser: The compiled parser. Is an object of the class SimpleParser in nomadcore.simple_parser.py.
        """
        self.parser = parser

        # allows to reset values if the same superContext is used to parse different files
        # self.dos_energies = None
        # self.dos_values = None

        # allows to reset values if the same superContext is used to parse different files
        self.initialize_values()


    def initialize_values(self):
        """Initializes the values of certain variables.

        This allows a consistent setting and resetting of the variables,
        when the parsing starts and when a section_run closes.
        """
        self.masses = {}


    # def onClose_section_dos(self, backend, gIndex, section):
    #     """Trigger called when section_dos is closed.
    #
    #     Store the parsed values and write them if writeMetaData is True.
    #     """
    #     # extract energies
    #     dos_energies = section['fhi_aims_dos_energy']
    #     if dos_energies is not None:
    #         pass
    #         # self.dos_energies = np.asarray(dos_energies)
    #     # extract dos values
    #     dos_values = []
    #     if section['fhi_aims_dos_value_string'] is not None:
    #         for string in section['fhi_aims_dos_value_string']:
    #             strings = string.split()
    #             dos_values.append(map(float, strings))
    #     if dos_values:
    #         # need to transpose array since its shape is [number_of_spin_channels,n_dos_values] in the metadata
    #         pass
    #         # self.dos_values = np.transpose(np.asarray(dos_values))
    #     # write metadata only if values were found for both quantities
    #     if self.writeMetaData:
    #         if dos_energies is not None and dos_values:
    #             pass
    #             # backend.addArrayValues('dos_energies', self.dos_energies)
    #             # backend.addArrayValues('dos_values', self.dos_values)

    def onClose_section_run(self, backend, gIndex, section):
        pass

    def onClose_section_topology(self, backend, gIndex, section):

        ####################################################################################################################################################################################################################################
        ####### RETRIEVING TOPOLOGY AND FORCE FIELD INFO FROM ANCILLARY PARSERS ############################################################################################################################################################
        ####################################################################################################################################################################################################################################

        topo_list = []
        if  section['x_lammps_data_topo_list_store']:
            for line in section['x_lammps_data_topo_list_store']:
                topo_list.append(line.split())

            topo_list.sort(key=lambda x: int(x[0]))  # ordering the atom list (important if the data file is generated from a binary restart file)

        # collecting atomic masses
        charge_dict, charge_list, mass_dict, mass_list, mass_xyz, new_mass_list, atomLabelling  = self.readChargeAndMass(topo_list, section)
        mass_dict = sorted(list(mass_dict.items()), key=operator.itemgetter(0))
        charge_dict = sorted(list(charge_dict.items()), key=operator.itemgetter(0)) # ordering atomic partial charges

        # return values of the original function
        self.charge_dict = charge_dict
        self.charge_list = charge_list
        self.mass_dict = mass_dict
        self.mass_list = mass_list
        self.mass_xyz = mass_xyz
        self.new_mass_list = new_mass_list
        self.atomLabelling = atomLabelling


        updateAtomTypes = []  ### to account for atoms with the same type in the topology, but with different partial charges
        ### see function readChargeAndMass in LAMMPSParserData
        if new_mass_list:
            for i in range(len(mass_list)):
                updateAtomTypes.append([ new_mass_list[i][0], mass_list[i][0] ])
        else:
            pass
            ###

        self.updateAtomTypes = updateAtomTypes


        # collecting information defining different molecules
        # LEGEND:
        # moleculeTypeInfo     --> for each molecule type we store [ molecule type index, atom type each atom in the molecule,
        #                                                            absolute atom index for the first occurence of this molecule type in the topology ]
        # moleculeInfo         --> for each molecule we store      [ absoulute molecule index, molecule type, absolute molecular atom index within the topology,
        #                                                            molecular atom index within the molecule ]
        # moleculeInfoResolved --> for each topology atom we store [ absolute atom index within the topology, molecule index,
        #                                                            molecule type index, molecular atom index within the molecule ]


        bond_list = []    #  LIST STORING ALL BONDS
        if section['x_lammps_data_bond_list_store']:
            for line in section['x_lammps_data_bond_list_store']:
                bond_list.append(line.split())

            bond_list.sort(key=lambda x: int(x[0]))


        updateAtomTypes = []  ### to account for atoms with the same type in the topology, but with different partial charges
        ### see function readChargeAndMass in LAMMPSParserData
        if new_mass_list:
            for i in range(len(mass_list)):
                updateAtomTypes.append([ new_mass_list[i][0], mass_list[i][0] ])
        else:
            pass
            ###


        # collecting information defining different molecules
        # LEGEND:
        # moleculeTypeInfo     --> for each molecule type we store [ molecule type index, atom type each atom in the molecule,
        #                                                            absolute atom index for the first occurence of this molecule type in the topology ]
        # moleculeInfo         --> for each molecule we store      [ absoulute molecule index, molecule type, absolute molecular atom index within the topology,
        #                                                            molecular atom index within the molecule ]
        # moleculeInfoResolved --> for each topology atom we store [ absolute atom index within the topology, molecule index,
        #                                                            molecule type index, molecular atom index within the molecule ]

        moleculeTypeInfo, moleculeInfo, moleculeInfoResolved = self.assignMolecules(topo_list, bond_list, section)

        # return values of assignMolecules
        self.moleculeTypeInfo = moleculeTypeInfo
        self.moleculeInfo = moleculeInfo
        self.moleculeInfoResolved = moleculeInfoResolved
        ###

        # # collecting list of force field functional styles
        # list_of_styles = readStyles()
        # pairFunctionalAndCutOff = list_of_styles.get('pair_style')  ### pair interactions style LAMMPS name + cutoff
        # pairFunctional = pairFunctionalAndCutOff[0]                 ### pair interactions style LAMMPS name only
        # bondFunctional = list_of_styles.get('bond_style')
        # angleFunctional = list_of_styles.get('angle_style')
        # dihedralFunctional = list_of_styles.get('dihedral_style')
        ###

        # collecting covalent bond definitions
        bond_dict, bondTypeList, bond_interaction_atoms  = self.assignBonds(topo_list, bond_list, updateAtomTypes, section)
        bond_dict = sorted(list(bond_dict.items()), key=operator.itemgetter(0))
        bd_types = len(bond_dict)

        self.bond_list = bond_list
        self.bond_dict = bond_dict
        self.bondTypeList = bondTypeList
        self.bond_interaction_atoms = bond_interaction_atoms
        self.bd_types = bd_types
        ###

        angle_list = []    #  LIST STORING ALL ANGLES
        if section['x_lammps_data_angle_list_store']:
            for line in section['x_lammps_data_angle_list_store']:
                angle_list.append(line.split())

            angle_list.sort(key=lambda x: int(x[0]))

        # collecting bond angles definitions
        angle_dict, angleTypeList, angle_interaction_atoms  = self.assignAngles(angle_list, topo_list, updateAtomTypes)
        angle_dict = sorted(list(angle_dict.items()), key=operator.itemgetter(0))
        ag_types = len(angle_dict)
        ###

        self.angle_list = angle_list
        self.angle_dict = angle_dict
        self.angleTypeList = angleTypeList
        self.angle_interaction_atoms = angle_interaction_atoms
        self.ag_types = ag_types

        dihedral_list = []    #  LIST STORING ALL DIHEDRALS
        if section['x_lammps_data_dihedral_list_store']:
            for line in section['x_lammps_data_dihedral_list_store']:
                dihedral_list.append(line.split())

            dihedral_list.sort(key=lambda x: int(x[0]))


        # collecting dihedral angles definitions
        dihedral_dict, dihedralTypeList, dihedral_interaction_atoms  = self.assignDihedrals(topo_list, dihedral_list, updateAtomTypes)
        dihedral_dict = sorted(list(dihedral_dict.items()), key=operator.itemgetter(0))
        dh_types = len(dihedral_dict)
        ###


        self.dihedral_list = dihedral_list
        self.dihedral_dict = dihedral_dict
        self.dihedralTypeList = dihedralTypeList
        self.dihedral_interaction_atoms = dihedral_interaction_atoms
        self.dh_types = dh_types

        dihedral_coeff_list = []
        if section['x_lammps_data_dihedral_coeff_list_store']:
            for line in section['x_lammps_data_dihedral_coeff_list_store']:
                dihedral_list.append(line.split())

            dihedral_list.sort(key=lambda x: int(x[0]))


        # # NUMBER OF BONDS (bd_count)
        # SM(r"\s(?P<x_lammps_data_bd_count_store>{0})\sbond".format(sformat.int)),
        #
        # # NUMBER OF ANGLES (ag_count)
        # SM(r"\s(?P<x_lammps_data_ag_count_store>{0})\sangles".format(sformat.int)),
        #
        # # NUMBER OF DIHEDRALS (dh_count)
        # SM(r"\s(?P<x_lammps_data_dh_count_store>{0})\sdihedrals".format(sformat.int)),
        # SM(r"\s(?P<x_lammps_dummy>{0})\simpropers".format(sformat.int)),
        #
        #
        #
        # # NUMBER OF BOND TYPES
        # SM(r"\s(?P<x_lammps_data_bd_types_store>{0})\sbond types".format(sformat.int)),
        #
        # # NUMBER OF ANGLE TYPES (ag_types)
        # SM(r"\s(?P<x_lammps_dummy>{0})\sangle types".format(sformat.int)),
        #
        # # NUMBER OF DIHEDRAL TYPES (dh_types)
        pass

    ########################################################################################################################
    # moleculeTypeInfo, moleculeInfo, moleculeInfoResolved = self.assignMolecules(backend, gIndex, section)
    ########################################################################################################################
    def assignMolecules(self,topo_list, bond_list, section):  # FINDING INDIVIDUAL MOLECULES FROM BONDING PATTERN

        store = []
        for i in range(len(bond_list)):
            at1   = int(bond_list[i][2])  # atom 1 index
            at2   = int(bond_list[i][3])  # atom 2 index
            store.append([at1, at2])


        storeLine = []
        for i in range(len(bond_list)):
            at1   = int(bond_list[i][2])  # atom 1 index
            at2   = int(bond_list[i][3])  # atom 2 index
            storeLine.append(at1)
            storeLine.append(at2)
            #print len(store)

        ### TOPOLOGY'S ATOM TYPE PATTERN ###
        atomTypeInTopology = []
        for i in range(len(topo_list)):
            temp = int(topo_list[i][2])
            atomTypeInTopology.append(temp) # Atom type pattern throughout topology

        topologyPattern = " ".join([ str(x) for x in atomTypeInTopology])
        #print topologyPattern
        ############

        #####################################################################
        ###### FIND ALL MOLECULE TYPES AND ALL INDIVIDUAL MOLECULES #########
        #####################################################################

        goon             = True # looping over all covalent bonds
        nextMolecule     = 0
        moleculeId       = 0    # molecular index
        moleculePattern  = str  # atom type pattern as string (useful for matching/counting)
        moleculeInfo     = []   # list containing molecule info


        while goon == True:
            atomIndexInMolecule = [] # list storing the index of atoms within that molecule
            atomTypeInMolecule  = [] # list storing the type of atoms within that molecule

            atomPositionInMolecule = [] # list storing the molecular atom index, example: OHH -> 123
            atomCount = 0

            for i in range(nextMolecule, len(store)):
                atomIndexInMolecule.append(store[i][0])
                atomIndexInMolecule.append(store[i][1])

                try:
                    if store[i+1][0] not in atomIndexInMolecule and store[i+1][1] not in atomIndexInMolecule:
                        moleculeId += 1
                        #print moleculeId
                        break
                except IndexError:
                    moleculeId += 1
                    #print moleculeId
                    break


            atomIndexToTopology = atomIndexInMolecule
            nextMolecule += len(atomIndexToTopology) // 2 ######## NOTA BENE ## I will start from this  to find the next molecule in the list "store"
            #print nextMolecule

            atomIndexInMolecule = sorted(list(set(atomIndexInMolecule)))  # clear duplicates and return the list of atom indexes in the molecule

            for i in atomIndexInMolecule:
                temp = int(topo_list[i-1][2])
                atomTypeInMolecule.append(temp)

                atomCount += 1
                atomPositionInMolecule.append(atomCount)

            moleculePattern = " ".join([ str(x) for x in atomTypeInMolecule])

            newMolecule = [ moleculeId, atomIndexInMolecule, atomTypeInMolecule, atomPositionInMolecule ] # storing molecule information
            moleculeInfo.append(newMolecule)

            #print atomIndexInMolecule
            #print atomTypeInMolecule

            if nextMolecule == len(store):
                goon = False

            #### Storing molecule info for each molecule type (the size of moleculeTypeInfo is the number of molecule types)
        moleculeTypeInfo = []
        ghost = []
        for line in moleculeInfo:
            seen = line[2]
            temp = [line[0], line[2], line[1]]

            if seen not in ghost:
                ghost.append(seen)
                moleculeTypeInfo.append(temp)
            #################

        for i in range(len(moleculeTypeInfo)):
            for j in range(len(moleculeInfo)):

                if moleculeTypeInfo[i][1] == moleculeInfo[j][2]:
                    moleculeInfo[j].insert(1, i+1)   ### moleculeInfo contains: [ moleculeId, moleculeType, atomIndexInMolecule,
                    ###                          atomTypeInMolecule, atomPositionInMolecule ]

        atomPositionInMoleculeList = []
        for i in range(len(moleculeInfo)):
            atomPositionInMoleculeList = atomPositionInMoleculeList + moleculeInfo[i][4] # complete list storing the molecular atom index, example: OHHOHHOHH -> 123123123

        # NUMBER OF ATOMS (at_count)
        at_count = int(section['number_of_topology_atoms'][0])
        self.at_count = at_count

        moleculeInfoResolved = []
        for i in range(0, at_count):
            for j in range(len(moleculeInfo)):

                if i+1 in moleculeInfo[j][2]:
                    moleculeInfoResolved.append([ i+1, moleculeInfo[j][0], moleculeInfo[j][1], atomPositionInMoleculeList[i] ])


        # print moleculeTypeInfo
        return (moleculeTypeInfo, moleculeInfo, moleculeInfoResolved)



        # collecting covalent bond definitions
        # bond_dict, bondTypeList, bond_interaction_atoms  = assignBonds(updateAtomTypes)
        # def assignBonds(updateAtomTypes):  # ASSIGNING COVALENT BOND TO ITS ATOM PAIR

        bond_ass_d = {}
        bond_ass = []
        for line in bond_list:
            nr    = line[0]
            index = line[1]
            at1   = line[2]
            at2   = line[3]

            store = { nr : [index, at1, at2] }

            if index not in bond_ass:
                bond_ass.append(index)
                bond_ass_d.update(store)

        bond_pairs = []
        for key, value in bond_ass_d.items():
            temp = value
            bond_pairs.append(temp)

        bond_dict = {}
        bondTypeList = []
        for line in bond_pairs:
            ind = int(line[0])
            at1 = int(line[1])
            at2 = int(line[2])
            bondTypeList.append(ind)

            a = topo_list[at1-1][2]
            b = topo_list[at2-1][2]
            bd = { ind : [int(a), int(b)] }
            bond_dict.update(bd)
        #bond_dict = { "Bond assignement (index, at_type1, at_type1)" : bond_dict }

        bondTypeList = sorted(bondTypeList)

        bond_interaction_atoms = []
        for i in bondTypeList:
            for line in bond_list:

                if int(line[1]) == i:
                    #count += 1
                    hit = [ int(line[1]), int(line[2]), int(line[3]) ]
                    bond_interaction_atoms.append(hit)


        ##################
        if updateAtomTypes:

            bond_ass   = []
            for line in bond_list:

                at1   = int(line[2])
                at2   = int(line[3])

                type_at1 = int(topo_list[at1-1][2])
                type_at2 = int(topo_list[at2-1][2])

                store = [ type_at1, type_at2 ]
                store = sorted(store)

                if store not in bond_ass:
                    bond_ass.append(store)

                bond_ass = sorted(bond_ass)

            bond_ass_eq = copy.deepcopy(bond_ass)

            for line in updateAtomTypes:
                if line[0] != line[1]:

                    for i in range(len(bond_ass_eq)):
                        for j in range(2):
                            if bond_ass_eq[i][j] == line[1]:
                                bond_ass_eq[i][j] = line[0]

            # for line in bond_list:
            #     ind1 = int(line[2])
            #     ind2 = int(line[3])
            #     to_atom_type = sorted( [ int(topo_list[ind1-1][2]), int(topo_list[ind2-1][2]) ] )
            #
            #     new_bond_index = bond_ass.index(to_atom_type) + 1
            #
            #     new_bond_list_line = [ int(line[0]), new_bond_index, ind1, ind2 ]

            new_bond_list =[]
            for line in bond_list:
                bd   = int(line[1])
                ind1 = int(line[2])
                ind2 = int(line[3])
                # to_atom_type = sorted( [ int(topo_list[ind1-1][2]), int(topo_list[ind2-1][2]) ] )

                new_bond_list_line = [ int(line[0]), bd, int(topo_list[ind1-1][2]), int(topo_list[ind2-1][2]) ]
                # print new_bond_list_line
                new_bond_list.append(new_bond_list_line)

            bond_list_unique = []
            for bond in bond_ass:
                for line in new_bond_list:

                    temp = [ line[1], [ line[2], line[3] ] ]
                    if temp not in bond_list_unique and [temp[0], temp[1][::-1]] not in bond_list_unique:   # VIP
                        bond_list_unique.append(temp)

            updated_bond_list = list(sorted(bond_list_unique))

            bond_dict = {}
            for bond in updated_bond_list:
                bond_dict.setdefault(bond[0], [])
                bond_dict[bond[0]].append(bond[1])

                # for bond in updated_bond_list:
                #     temp = { bond[0] : bond[1] }
                #     new_bond_dict.update(temp)

        # return (bond_dict, bondTypeList, bond_interaction_atoms)

        # bond_dict = sorted(list(bond_dict.items()), key=operator.itemgetter(0))
        #
        # list_of_bonds = readBonds(bondFunctional)
        # list_of_bonds = sorted(list(list_of_bonds.items()), key=operator.itemgetter(0))
        # bd_types = len(bond_dict)
        ###

    def readChargeAndMass(self, topo_list, section):  ### here we record atomic masses and partial charges
        charge_dict   = {}
        charge_list   = []
        mass_dict     = {}
        mass_list     = []
        mass_xyz      = []
        new_mass_list = []


        for line in topo_list:
            index = int(line[2])
            charge = float(line[3])
            seen = [index, charge]
            store = { index : charge }

            if seen not in charge_list:
                charge_list.append(seen)
                charge_dict.update(store)


        # NUMBER OF ATOMS (at_count)
        at_count = int(section['number_of_topology_atoms'][0])

        # NUMBER OF ATOM TYPES
        at_types = section['x_lammps_data_at_types_store'][0]


        switch = False
        if at_types == len(charge_list):

            switch = True

            masses = section['x_lammps_masses_store']
            for data in masses:
                mass = data.split()
                index = int(mass[0])
                val   = float(mass[1])
                val1  = int(val/2)     # REQUIRES DOUBLE CHECK  (calculates atomic number for labelling purposes)

                # create list
                store = [index, val]
                mass_xyz.append(val1)
                mass_list.append(store)

                # create dictionary
                masses = { index : val }
                mass_dict.update(masses)


            for i in range(0,len(mass_xyz)):
                if mass_xyz[i] == 0:
                    mass_xyz[i] = 1


            xyz_file = []     # WRITE AN XYZ FILE FROM LAMMPS TOPOLOGY DATA
            atomLabelling = []
            xyz_file.append([at_count])
            xyz_file.append([' '])
            for line in topo_list:
                index = int(line[2])
                xyz_line = [mass_xyz[index-1], float(line[4]), float(line[5]),  float(line[6])]
                xyz_file.append(xyz_line)
                atomLabelling.append(xyz_line)

            # mode = 'w'
            # if sys.version_info.major < 3:
            #     mode += 'b'
            #
            # with open(os.path.normpath(os.path.join(os.path.dirname(os.path.abspath(sys.argv[1])), 'generated_from_data_file.xyz')), mode) as xyz:
            #     xyz.writelines('  '.join(str(j) for j in i) + '\n' for i in xyz_file)    # WRITE XYZ ATOMIC NUMBER AND COORDINATES


        #### A SINGLE ATOM TYPE MIGHT HAVE MORE THAN ONE CHARGE (E.G. CARBON IN CH3, CH2, CH, ...)
        #### WITH THE CODE BELOW, WE CREATE A NEW ATOM TYPE WHENEVER A SINGLE ATOM TYPE HAS MORE THAN ONE ASSOCIATED PARTIAL CHARGE
        if switch == False:


            masses = section['x_lammps_masses_store']
            for data in masses:
                mass = data.split()

                index = int(mass[0])
                val   = float(mass[1])
                val1  = int(val/2)     # REQUIRES DOUBLE CHECK


                # create list
                store = [index, val]
                mass_xyz.append(val1)
                mass_list.append(store)

                # create dictionary
                masses = { index : val }
                mass_dict.update(masses)

            #print mass_list

            new_mass_list = []
            for type in charge_list:
                index = type[0]-1
                # print index
                new_mass_list.append([type[0], mass_list[index][1]])

            #print new_mass_list
            mass_list = list(new_mass_list)

            #########
            for type in mass_list:
                temp = int(type[1]/2)
                mass_xyz.append(temp)

            for i in range(0,len(mass_xyz)):
                if mass_xyz[i] == 0:
                    mass_xyz[i] = 1
            #########

            for i in range(len(charge_list)):
                new_mass_list[i] = [ charge_list[i][0], mass_list[i][1] ]
                mass_list[i][0] = i + 1
                mass_dict.update({ mass_list[i][0] : mass_list[i][1] })
                charge_list[i][0] = i + 1
                charge_dict.update({ charge_list[i][0] : charge_list[i][1] })

                #print mass_list

            topo_list_new = []
            for line in topo_list:
                topo_list_new.append(line)

            for type in charge_list:
                for i in range(len(topo_list_new)):

                    if type[1] == float(topo_list_new[i][3]):
                        topo_list_new[i][2] = str(type[0])

            topo_list_new.sort(key=lambda x: int(x[0]))

            #print topo_list_new

            xyz_file = []     # WRITE AN XYZ FILE FROM LAMMPS TOPOLOGY DATA
            atomLabelling = []
            xyz_file.append([at_count])
            xyz_file.append([' '])
            for line in topo_list_new:
                index = int(line[2])
                xyz_line = [mass_xyz[index-1], float(line[4]), float(line[5]),  float(line[6])]
                xyz_file.append(xyz_line)
                atomLabelling.append(xyz_line)

            # mode = 'w'
            # if sys.version_info.major < 3:
            #     mode += 'b'
            #
            # with open(os.path.normpath(os.path.join(os.path.dirname(os.path.abspath(sys.argv[1])), 'generated_from_data_file.xyz')), mode) as xyz:
            #     xyz.writelines('  '.join(str(j) for j in i) + '\n' for i in xyz_file)    # WRITE XYZ ATOMIC NUMBER AND COORDINATES

        return charge_dict, charge_list, mass_dict, mass_list, mass_xyz, new_mass_list, atomLabelling


    ########################################################################################################################
    def assignBonds(self, topo_list, bond_list, updateAtomTypes, section):  # ASSIGNING COVALENT BOND TO ITS ATOM PAIR

        bond_ass_d = {}
        bond_ass = []
        for line in bond_list:
            nr    = line[0]
            index = line[1]
            at1   = line[2]
            at2   = line[3]

            store = { nr : [index, at1, at2] }

            if index not in bond_ass:
                bond_ass.append(index)
                bond_ass_d.update(store)

        bond_pairs = []
        for key, value in bond_ass_d.items():
            temp = value
            bond_pairs.append(temp)

        bond_dict = {}
        bondTypeList = []
        for line in bond_pairs:
            ind = int(line[0])
            at1 = int(line[1])
            at2 = int(line[2])
            bondTypeList.append(ind)

            a = topo_list[at1-1][2]
            b = topo_list[at2-1][2]
            bd = { ind : [int(a), int(b)] }
            bond_dict.update(bd)
        #bond_dict = { "Bond assignement (index, at_type1, at_type1)" : bond_dict }

        bondTypeList = sorted(bondTypeList)

        bond_interaction_atoms = []
        for i in bondTypeList:
            for line in bond_list:

                if int(line[1]) == i:
                    #count += 1
                    hit = [ int(line[1]), int(line[2]), int(line[3]) ]
                    bond_interaction_atoms.append(hit)


        ##################
        if updateAtomTypes:

            bond_ass   = []
            for line in bond_list:

                at1   = int(line[2])
                at2   = int(line[3])

                type_at1 = int(topo_list[at1-1][2])
                type_at2 = int(topo_list[at2-1][2])

                store = [ type_at1, type_at2 ]
                store = sorted(store)

                if store not in bond_ass:
                    bond_ass.append(store)

                bond_ass = sorted(bond_ass)

            bond_ass_eq = copy.deepcopy(bond_ass)

            for line in updateAtomTypes:
                if line[0] != line[1]:

                    for i in range(len(bond_ass_eq)):
                        for j in range(2):
                            if bond_ass_eq[i][j] == line[1]:
                                bond_ass_eq[i][j] = line[0]

            # for line in bond_list:
            #     ind1 = int(line[2])
            #     ind2 = int(line[3])
            #     to_atom_type = sorted( [ int(topo_list[ind1-1][2]), int(topo_list[ind2-1][2]) ] )
            #
            #     new_bond_index = bond_ass.index(to_atom_type) + 1
            #
            #     new_bond_list_line = [ int(line[0]), new_bond_index, ind1, ind2 ]

            new_bond_list =[]
            for line in bond_list:
                bd   = int(line[1])
                ind1 = int(line[2])
                ind2 = int(line[3])
                # to_atom_type = sorted( [ int(topo_list[ind1-1][2]), int(topo_list[ind2-1][2]) ] )

                new_bond_list_line = [ int(line[0]), bd, int(topo_list[ind1-1][2]), int(topo_list[ind2-1][2]) ]
                # print new_bond_list_line
                new_bond_list.append(new_bond_list_line)

            bond_list_unique = []
            for bond in bond_ass:
                for line in new_bond_list:

                    temp = [ line[1], [ line[2], line[3] ] ]
                    if temp not in bond_list_unique and [temp[0], temp[1][::-1]] not in bond_list_unique:   # VIP
                        bond_list_unique.append(temp)

            updated_bond_list = list(sorted(bond_list_unique))

            bond_dict = {}
            for bond in updated_bond_list:
                bond_dict.setdefault(bond[0], [])
                bond_dict[bond[0]].append(bond[1])

                # for bond in updated_bond_list:
                #     temp = { bond[0] : bond[1] }
                #     new_bond_dict.update(temp)

        return (bond_dict, bondTypeList, bond_interaction_atoms)

    # bond_dict, bondTypeList, bond_interaction_atoms = assignBonds(updateAtomTypes)

    ########################################################################################################################
    def assignAngles(self, angle_list, topo_list, updateAtomTypes):  # ASSIGNING ANGLE TO ITS ATOM TRIPLET

        angle_ass_d = {}
        angle_ass = []
        for line in angle_list:
            nr    = line[0]
            index = line[1]
            at1   = line[2]
            at2   = line[3]
            at3   = line[4]

            store = { nr : [index, at1, at2, at3] }

            if index not in angle_ass:
                angle_ass.append(index)
                angle_ass_d.update(store)

        angle_triplets = []
        for key, value in angle_ass_d.items():
            temp = value
            angle_triplets.append(temp)

        angle_dict = {}
        angleTypeList = []
        for line in angle_triplets:
            ind = int(line[0])
            at1 = int(line[1])
            at2 = int(line[2])
            at3 = int(line[3])
            angleTypeList.append(ind)

            a = topo_list[at1-1][2]
            b = topo_list[at2-1][2]
            c = topo_list[at3-1][2]
            ag = { ind : [int(a), int(b), int(c)] }
            angle_dict.update(ag)
        #angle_dict = { "Angle assignement (index, at_type1, at_type1, at_type3)" : angle_dict }

        angleTypeList = sorted(angleTypeList)

        angle_interaction_atoms = []
        for i in angleTypeList:
            for line in angle_list:

                if int(line[1]) == i:
                    #count += 1
                    hit = [ int(line[1]), int(line[2]), int(line[3]), int(line[4]) ]
                    angle_interaction_atoms.append(hit)


        ##################
        if updateAtomTypes:

            angle_ass   = []
            for line in angle_list:

                at1   = int(line[2])
                at2   = int(line[3])
                at3   = int(line[4])

                type_at1 = int(topo_list[at1-1][2])
                type_at2 = int(topo_list[at2-1][2])
                type_at3 = int(topo_list[at3-1][2])

                store = [ type_at1, type_at2, type_at3 ]
                #store = sorted(store)

                if store not in angle_ass:
                    angle_ass.append(store)

            angle_ass = sorted(angle_ass)

            angle_ass_eq = copy.deepcopy(angle_ass)

            for line in updateAtomTypes:
                if line[0] != line[1]:

                    for i in range(len(angle_ass_eq)):
                        for j in range(3):
                            if angle_ass_eq[i][j] == line[1]:
                                angle_ass_eq[i][j] = line[0]

            new_angle_list =[]
            for line in angle_list:
                bd   = int(line[1])
                ind1 = int(line[2])
                ind2 = int(line[3])
                ind3 = int(line[4])

                new_angle_list_line = [ int(line[0]), bd, int(topo_list[ind1-1][2]), int(topo_list[ind2-1][2]), int(topo_list[ind3-1][2]) ]
                new_angle_list.append(new_angle_list_line)

            angle_list_unique = []
            for angle in angle_ass:
                for line in new_angle_list:

                    temp = [ line[1], [ line[2], line[3], line[4] ] ]
                    if temp not in angle_list_unique and [temp[0], temp[1][::-1]] not in angle_list_unique:   # VIP
                        angle_list_unique.append(temp)

            updated_angle_list = list(sorted(angle_list_unique))

            angle_dict = {}
            for angle in updated_angle_list:
                angle_dict.setdefault(angle[0], [])
                angle_dict[angle[0]].append(angle[1])

                # print angle_ass, '########'
                # print angle_dict, '#######'

        return (angle_dict, angleTypeList, angle_interaction_atoms)

    #angle_dict, angleTypeList, angle_interaction_atoms = assignAngles()

        ########################################################################################################################
    def assignDihedrals(self, topo_list, dihedral_list, updateAtomTypes):  # ASSIGNING DIHEDRAL TO ITS ATOM QUARTET

        dihedral_ass_d = {}
        dihedral_ass = []
        for line in dihedral_list:
            nr    = line[0]
            index = line[1]
            at1   = line[2]
            at2   = line[3]
            at3   = line[4]
            at4   = line[5]

            store = { nr : [index, at1, at2, at3, at4] }

            if index not in dihedral_ass:
                dihedral_ass.append(index)
                dihedral_ass_d.update(store)

        dihedral_quartets = []
        for key, value in dihedral_ass_d.items():
            temp = value
            dihedral_quartets.append(temp)

        dihedral_dict = {}
        dihedralTypeList = []
        for line in dihedral_quartets:
            ind = int(line[0])
            at1 = int(line[1])
            at2 = int(line[2])
            at3 = int(line[3])
            at4 = int(line[4])
            dihedralTypeList.append(ind)

            a = topo_list[at1-1][2]
            b = topo_list[at2-1][2]
            c = topo_list[at3-1][2]
            d = topo_list[at4-1][2]
            dh = { ind : [int(a), int(b), int(c), int(d)] }
            dihedral_dict.update(dh)
        #dihedral_dict = { "Dihedral assignement (index, at_type1, at_type1, at_type3, at_type4)" : dihedral_dict }

        dihedralTypeList = sorted(dihedralTypeList)

        dihedral_interaction_atoms = []
        for i in dihedralTypeList:
            for line in dihedral_list:

                if int(line[1]) == i:
                    #count += 1
                    hit = [ int(line[1]), int(line[2]), int(line[3]), int(line[4]), int(line[5])  ]
                    dihedral_interaction_atoms.append(hit)


        ##################
        if updateAtomTypes:

            dihedral_ass   = []
            for line in dihedral_list:

                at1   = int(line[2])
                at2   = int(line[3])
                at3   = int(line[4])
                at4   = int(line[5])


                type_at1 = int(topo_list[at1-1][2])
                type_at2 = int(topo_list[at2-1][2])
                type_at3 = int(topo_list[at3-1][2])
                type_at4 = int(topo_list[at4-1][2])

                store = [ type_at1, type_at2, type_at3, type_at4 ]
                #store = sorted(store)

                if store not in dihedral_ass:
                    dihedral_ass.append(store)

            dihedral_ass = sorted(dihedral_ass)

            dihedral_ass_eq = copy.deepcopy(dihedral_ass)

            for line in updateAtomTypes:
                if line[0] != line[1]:

                    for i in range(len(dihedral_ass_eq)):
                        for j in range(4):
                            if dihedral_ass_eq[i][j] == line[1]:
                                dihedral_ass_eq[i][j] = line[0]

            new_dihedral_list =[]
            for line in dihedral_list:
                bd   = int(line[1])
                ind1 = int(line[2])
                ind2 = int(line[3])
                ind3 = int(line[4])
                ind4 = int(line[5])

                new_dihedral_list_line = [ int(line[0]), bd, int(topo_list[ind1-1][2]), int(topo_list[ind2-1][2]), int(topo_list[ind3-1][2]), int(topo_list[ind4-1][2]) ]
                new_dihedral_list.append(new_dihedral_list_line)

            dihedral_list_unique = []
            for dihedral in dihedral_ass:
                for line in new_dihedral_list:

                    temp = [ line[1], [ line[2], line[3], line[4], line[5] ] ]
                    if temp not in dihedral_list_unique and [temp[0], temp[1][::-1]] not in dihedral_list_unique:   # VIP
                        dihedral_list_unique.append(temp)

            updated_dihedral_list = list(sorted(dihedral_list_unique))

            # print updated_dihedral_list, "#######"

            dihedral_dict = {}
            for dihedral in updated_dihedral_list:
                dihedral_dict.setdefault(dihedral[0], [])
                dihedral_dict[dihedral[0]].append(dihedral[1])

                # print dihedral_ass, '########'
                # print dihedral_dict, '#######'

        return (dihedral_dict, dihedralTypeList, dihedral_interaction_atoms)

        #dihedral_dict, dihedralTypeList, dihedral_interaction_atoms = assignDihedrals()

        #print dihedralTypeList
        #print dihedral_interaction_atoms


    ########################################################################################################################
    def readBondsFromData(self, bond_list, bondFunctional):

        toMass     = self.converter.ratioMass
        toDistance = self.converter.ratioDistance
        toTime     = self.converter.ratioTime
        toEnergy   = self.converter.ratioEnergy
        toVelocity = self.converter.ratioVelocity
        toForce    = self.converter.ratioForce
        toTorque   = self.converter.ratioTorque
        toTemp     = self.converter.ratioTemp
        toPress    = self.converter.ratioPress
        toDVisc    = self.converter.ratioDVisc
        toCharge   = self.converter.ratioCharge
        toDipole   = self.converter.ratioDipole
        toEField   = self.converter.ratioEField
        toDensity  = self.converter.ratioDensity


        list_of_bonds={}
        for line in bond_list:

            if bondFunctional == "harmonic":
                index1 = int(line[0])
                index2 = float(line[1])*(toEnergy/(toDistance)**2)
                index3 = float(line[2])*toDistance

                bond = [ index2, index3 ]
                bond_dict = {index1 : bond }
                list_of_bonds.update(bond_dict)


            if bondFunctional == "class2":   # COMPASS
                index1 = int(line[0])
                index2 = float(line[1])*toDistance
                index3 = float(line[2])*(toEnergy/(toDistance)**2)
                index4 = float(line[3])*(toEnergy/(toDistance)**3)
                index5 = float(line[4])*(toEnergy/(toDistance)**4)

                bond = [ index2, index3, index4, index5 ]
                bond_dict = {index1 : bond}
                list_of_bonds.update(bond_dict)


            if bondFunctional == "nonlinear":
                index1 = int(line[0])
                index2 = float(line[1])*toEnergy
                index3 = float(line[2])*toDistance
                index4 = float(line[3])*toDistance

                bond = [ index2, index3, index4 ]
                bond_dict = {index1 : bond}
                list_of_bonds.update(bond_dict)


            if bondFunctional == "morse":
                index1 = int(line[0])
                index2 = float(line[1])*toEnergy
                index3 = float(line[2])*(1/toDistance)
                index4 = float(line[3])*toDistance

                bond = [ index2, index3, index4 ]
                bond_dict = {index1 : bond}
                list_of_bonds.update(bond_dict)


        return list_of_bonds


    ########################################################################################################################
    def readAnglesFromData(self, angle_list, angleFunctional):

        toMass     = self.converter.ratioMass
        toDistance = self.converter.ratioDistance
        toTime     = self.converter.ratioTime
        toEnergy   = self.converter.ratioEnergy
        toVelocity = self.converter.ratioVelocity
        toForce    = self.converter.ratioForce
        toTorque   = self.converter.ratioTorque
        toTemp     = self.converter.ratioTemp
        toPress    = self.converter.ratioPress
        toDVisc    = self.converter.ratioDVisc
        toCharge   = self.converter.ratioCharge
        toDipole   = self.converter.ratioDipole
        toEField   = self.converter.ratioEField
        toDensity  = self.converter.ratioDensity

        toRadians = 0.0174533  # multiply to convert deg to rad


        list_of_angles={}
        for line in angle_list:

            if angleFunctional == "harmonic":
                index1 = int(line[0])
                index2 = float(line[1])*toEnergy
                index3 = float(line[2])*toRadians

                angle = [ index2, index3 ]
                angle_dict = {index1 : angle }
                list_of_angles.update(angle_dict)

            if angleFunctional == "class2":   # COMPASS
                pass

            if angleFunctional == "charmm":
                index1 = int(line[0])
                index2 = float(line[1])*toEnergy
                index3 = float(line[2])*toRadians
                index4 = float(line[3])*(toEnergy/(toDistance)**2)
                index5 = float(line[4])*toDistance

                angle = [ index2, index3, index4, index5 ]
                angle_dict = {index1 : angle }
                list_of_angles.update(angle_dict)

            if angleFunctional == "cosine":
                index1 = int(line[0])
                index2 = float(line[1])*toEnergy

                angle = [ index2 ]
                angle_dict = {index1 : angle }
                list_of_angles.update(angle_dict)

            if angleFunctional == "cosine/delta":
                index1 = int(line[0])
                index2 = float(line[1])*toEnergy
                index3 = float(line[2])*toRadians

                angle = [ index2, index3 ]
                angle_dict = {index1 : angle }
                list_of_angles.update(angle_dict)

            if angleFunctional == "cosine/periodic":   # DREIDING
                index1 = int(line[0])
                index2 = float(line[1])*toEnergy
                index3 = int(line[2])
                index4 = int(line[3])

                angle = [ index2, index3, index4 ]
                angle_dict = {index1 : angle }
                list_of_angles.update(angle_dict)


            if angleFunctional == "cosine/squared":
                index1 = int(line[0])
                index2 = float(line[1])*toEnergy
                index3 = float(line[2])*toRadians

                angle = [ index2, index3 ]
                angle_dict = {index1 : angle }
                list_of_angles.update(angle_dict)

        return list_of_angles


    ########################################################################################################################
    def readDihedralsFromData(self, dihedral_list, dihedralFunctional):

        toMass     = self.converter.ratioMass
        toDistance = self.converter.ratioDistance
        toTime     = self.converter.ratioTime
        toEnergy   = self.converter.ratioEnergy
        toVelocity = self.converter.ratioVelocity
        toForce    = self.converter.ratioForce
        toTorque   = self.converter.ratioTorque
        toTemp     = self.converter.ratioTemp
        toPress    = self.converter.ratioPress
        toDVisc    = self.converter.ratioDVisc
        toCharge   = self.converter.ratioCharge
        toDipole   = self.converter.ratioDipole
        toEField   = self.converter.ratioEField
        toDensity  = self.converter.ratioDensity

        toRadians = 0.0174533  # multiply to convert deg to rad

        list_of_dihedrals={}
        for line in dihedral_list:

            if dihedralFunctional == "harmonic":
                index1 = int(line[0])
                index2 = float(line[1])*toEnergy
                index3 = int(line[2])
                index4 = int(line[3])

                dihedral = [ index2, index3, index4 ]
                dihedral_dict = {index1 : dihedral }
                list_of_dihedrals.update(dihedral_dict)


            if dihedralFunctional == "class2":   # COMPASS
                pass


            if dihedralFunctional == "multi/harmonic":
                index1 = int(line[0])
                index2 = float(line[1])*toEnergy
                index3 = float(line[2])*toEnergy
                index4 = float(line[3])*toEnergy
                index5 = float(line[4])*toEnergy
                index6 = float(line[5])*toEnergy

                dihedral = [ index2, index3, index4, index5, index6 ]
                dihedral_dict = {index1 : dihedral }
                list_of_dihedrals.update(dihedral_dict)


            if dihedralFunctional == "charmm":
                index1 = int(line[0])
                index2 = float(line[1])*toEnergy
                index3 = int(line[2])
                index4 = float(line[3])*toRadians
                index5 = float(line[4])

                dihedral = [ index2, index3, index4, index5 ]
                dihedral_dict = {index1 : dihedral }
                list_of_dihedrals.update(dihedral_dict)


            if dihedralFunctional == "opls":   # OPLS aa
                index1 = int(line[0])
                index2 = float(line[1])*toEnergy
                index3 = float(line[2])*toEnergy
                index4 = float(line[3])*toEnergy
                index5 = float(line[4])*toEnergy

                dihedral = [ index2, index3, index4, index5 ]
                dihedral_dict = {index1 : dihedral }
                list_of_dihedrals.update(dihedral_dict)


            if dihedralFunctional == "helix":
                index1 = int(line[0])
                index2 = float(line[1])*toEnergy
                index3 = float(line[2])*toEnergy
                index4 = float(line[3])*toEnergy

                dihedral = [ index2, index3, index4 ]
                dihedral_dict = {index1 : dihedral }
                list_of_dihedrals.update(dihedral_dict)


        return list_of_dihedrals



    ########################################################################################################################
    def readPairCoeffFromData(self, lj_list, updateAtomTypes, pairFunctional):

        toMass     = self.converter.ratioMass
        toDistance = self.converter.ratioDistance
        toTime     = self.converter.ratioTime
        toEnergy   = self.converter.ratioEnergy
        toVelocity = self.converter.ratioVelocity
        toForce    = self.converter.ratioForce
        toTorque   = self.converter.ratioTorque
        toTemp     = self.converter.ratioTemp
        toPress    = self.converter.ratioPress
        toDVisc    = self.converter.ratioDVisc
        toCharge   = self.converter.ratioCharge
        toDipole   = self.converter.ratioDipole
        toEField   = self.converter.ratioEField
        toDensity  = self.converter.ratioDensity


        # this currently gather pair coefficients for pairFunctional reported in the list of strings below

        supportedLJFunct      = ['lj/cut', 'lj/cut/coul/cut', 'lj/cut/coul/long', 'lj/cut/coul/debye',
                                 'lj/cut/coul/dsf', 'lj/cut/coul/msm', 'lj/cut/tip4p/cut', 'lj/cut/tip4p/long']

        supportedCHARMMFunct  = ['lj/charmm/coul/charmm', 'lj/charmm/coul/charmm/implicit', 'lj/charmm/coul/long',
                                 'lj/charmm/coul/msm']

        supportedGROMACSFunct = ['lj/gromacs', 'lj/gromacs/coul/gromacs']


        list_of_ljs  = {}
        ljs_dict     = {}
        at_types_lj  = []
        index       = 0
        for line in lj_list:


            if pairFunctional in supportedLJFunct:
                index += 1
                atom1 = int(line[0])   # pair atom type 1
                atom2 = int(line[0])   # pair atom type 2
                eps   = float(line[1])*toEnergy     # epsilon
                sigma = float(line[2])*toDistance   # sigma

                coeff = [eps, sigma]

                for i in range(4, len(line)):  # if another float is present, it is the pair style cutoff(s) for this pair interaction
                    if "#" in line[i]:
                        break

                    try:
                        rad = float(line[i])*toDistance
                        coeff.append(rad)
                    except ValueError:
                        pass

                        # creat a list
                lj_coeff = [atom1, atom2, coeff]
                at_types_lj.append(lj_coeff)

                # create dictionaries
                lj_pair = { index : [atom1, atom2] }
                ljs_dict.update(lj_pair)

                lj_param = {index : coeff}
                list_of_ljs.update(lj_param)

            else:
                pass


            if pairFunctional in supportedCHARMMFunct:
                index += 1
                atom1 = int(line[0])   # pair atom type 1
                atom2 = int(line[0])   # pair atom type 2
                eps   = float(line[1])*toEnergy     # epsilon
                sigma = float(line[2])*toDistance   # sigma
                eps14   = float(line[3])*toEnergy     # epsilon 1-4
                sigma14 = float(line[4])*toDistance   # sigma   1-4

                coeff = [eps, sigma, eps14, sigma14]

                # creat a list
                lj_coeff = [atom1, atom2, coeff]
                at_types_lj.append(lj_coeff)

                # create dictionaries
                lj_pair = { index : [atom1, atom2] }
                ljs_dict.update(lj_pair)

                lj_param = {index : coeff}
                list_of_ljs.update(lj_param)

            else:
                pass


            if pairFunctional in supportedGROMACSFunct:
                index += 1
                atom1 = int(line[0])   # pair atom type 1
                atom2 = int(line[0])   # pair atom type 2
                eps   = float(line[1])*toEnergy     # epsilon
                sigma = float(line[2])*toDistance   # sigma
                inner = float(line[3])*toDistance   # inner sigma
                outer = float(line[4])*toDistance   # outer sigma

                coeff = [eps, sigma, inner, outer]

                # creat a list
                lj_coeff = [atom1, atom2, coeff]
                at_types_lj.append(lj_coeff)

                # create dictionaries
                lj_pair = { index : [atom1, atom2] }
                ljs_dict.update(lj_pair)

                lj_param = {index : coeff}
                list_of_ljs.update(lj_param)

            else:
                pass


        if updateAtomTypes:  # here I create pair styles including the new atom types (to account for atoms of the same type, but with different partial charges)

            for line in updateAtomTypes:
                if line[0] != line[1]:

                    list_of_ljs.setdefault(line[1], [])
                    try:
                        list_of_ljs[line[1]].append(list_of_ljs[line[0]][0])
                    except KeyError:
                        pass

                    try:
                        list_of_ljs[line[1]].append(list_of_ljs[line[0]][1])
                    except KeyError:
                        pass

                    ljs_dict.setdefault(line[1], [])
                    ljs_dict[line[1]].append(line[1])
                    ljs_dict[line[1]].append(line[1])

        return (list_of_ljs, ljs_dict)





def build_LammpsDataFileSimpleMatcher():
    """Builds the SimpleMatcher to parse the DATA file of LAMMPS.

    SimpleMatchers are called with 'SM (' as this string has length 4,
    which allows nice formating of nested SimpleMatchers in python.

    HEADERS = set([
        //'atoms',
        //'bonds',
        //'angles',
        //'dihedrals',
        //'impropers',
        //'atom types',
        //'bond types',
        //'angle types',
        //'dihedral types',
        //'improper types',
        //'extra bond per atom',
        //'extra angle per atom',
        //'extra dihedral per atom',
        //'extra improper per atom',
        //'extra special per atom',
        //'ellipsoids',
        //'lines',
        //'triangles',
        //'bodies',
        //'xlo xhi',
        //'ylo yhi',
        //'zlo zhi',
        //'xy xz yz',
    ])


    # Sections will all start with one of these words
    # and run until the next section title
    SECTIONS = set([
        //'Atoms',  # Molecular topology sections
        //'Velocities',
        //'Masses',
        //'Ellipsoids',
        //'Lines',
        //'Triangles',
        //'Bodies',
        //'Bonds',  # Forcefield sections
        //'Angles',
        //'Dihedrals',
        //'Impropers',
        //'Pair',
        //'Pair LJCoeffs',
        //'Bond Coeffs',
        //'Angle Coeffs',
        //'Dihedral Coeffs',
        //'Improper Coeffs',
        //'BondBond Coeffs',  # Class 2 FF sections
        //'BondAngle Coeffs',
        //'MiddleBondTorsion Coeffs',
        //'EndBondTorsion Coeffs',
        //'AngleTorsion Coeffs',
        //'AngleAngleTorsion Coeffs',
        //'BondBond13 Coeffs',
        //'AngleAngle Coeffs',
    ])


    Returns:
       SimpleMatcher that parses DATA file of FHI-LAMMPS.
    """
    atoms = SM(
        name='lammps-data-list-of-atoms',
        startReStr=r"\s*Atoms",
        repeats=True,
        subMatchers=[
            SM(
                r"\s*(?P<x_lammps_data_topo_list_store>{0}\s{0}\s{0}\s{1}\s{1}\s{1}\s{1})".format(sformat.int, sformat.float),
                repeats=True
            )
        ])

    velocities = SM(
        name='lammps-data-list-of-velocities',
        startReStr=r"\s*Velocities",
        repeats=True,
        subMatchers=[
            # SM(
            #     r"\s*(?P<x_lammps_data_topo_list_store>{0}\s{0}\s{0}\s{1}\s{1}\s{1}\s{1})".format(sformat.int, sformat.float),
            #     repeats=True
            # )
        ])

    masses = SM(
        name='lammps-data-list-of-masses',
        startReStr=r"\s*Masses",
        repeats=True,
        subMatchers=[
            SM(
                r"\s*(?P<x_lammps_masses_store>{0}\s{1})".format(sformat.int, sformat.float),
                repeats=True,
            )
        ])

    ellipsoids = SM(
        name='lammps-data-list-of-ellipsoids',
        startReStr=r"\s*Ellipsoids",
        repeats=True,
        subMatchers=[
            # SM(
            #     r"\s*(?P<x_lammps_data_topo_list_store>{0}\s{0}\s{0}\s{1}\s{1}\s{1}\s{1})".format(sformat.int, sformat.float),
            #     repeats=True
            # )
        ])



    lines = SM(
        name='lammps-data-list-of-lines',
        startReStr=r"\s*Lines",
        repeats=True,
        subMatchers=[
            # SM(
            #     r"\s*(?P<x_lammps_data_topo_list_store>{0}\s{0}\s{0}\s{1}\s{1}\s{1}\s{1})".format(sformat.int, sformat.float),
            #     repeats=True
            # )
        ])

    triangles = SM(
        name='lammps-data-list-of-triangles',
        startReStr=r"\s*Triangles",
        repeats=True,
        subMatchers=[
            # SM(
            #     r"\s*(?P<x_lammps_data_topo_list_store>{0}\s{0}\s{0}\s{1}\s{1}\s{1}\s{1})".format(sformat.int, sformat.float),
            #     repeats=True
            # )
        ])

    bodies = SM(
        name='lammps-data-list-of-bodies',
        startReStr=r"\s*Bodies",
        repeats=True,
        subMatchers=[
            # SM(
            #     r"\s*(?P<x_lammps_data_topo_list_store>{0}\s{0}\s{0}\s{1}\s{1}\s{1}\s{1})".format(sformat.int, sformat.float),
            #     repeats=True
            # )
        ])

    bonds = SM(
        name='lammps-data-list-of-bonds',
        startReStr=r"\s*Bonds",
        repeats=True,
        subMatchers=[
           SM(
               r"\s*(?P<x_lammps_data_bond_list_store>{0}\s{0}\s{0}\s{0})".format(sformat.int),
               repeats=True
           )
        ])

    angles = SM(
        name='lammps-data-list-of-bonds',
        startReStr=r"\s*Angles",
        repeats=True,
        subMatchers=[
            SM(
                r"\s*(?P<x_lammps_data_angle_list_store>{0}\s{0}\s{0}\s{0}\s{0})".format(sformat.int),
                repeats=True
            )
        ])

    dihedrals = SM(
        name='lammps-data-list-of-bonds',
        startReStr=r"\s*Dihedrals",
        repeats=True,
        subMatchers=[
            SM(
                r"\s*(?P<x_lammps_data_dihedral_list_store>{0}\s{0}\s{0}\s{0}(\s{0})+)".format(sformat.int),
                repeats=True
            )
        ])






    impropers = SM(
        name='lammps-data-list-of-bonds',
        startReStr=r"\s*Impropers",
        repeats=True,
        subMatchers=[
            # SM(
            #     r"\s*(?P<x_lammps_data_dihedral_coeff_list_store>{0}\s{0}\s{0}\s{0}\s{0})".format(sformat.int),
            #     repeats=True
            # )
        ])
    pair = SM(
        name='lammps-data-list-of-bonds',
        startReStr=r"\s*Pair",
        repeats=True,
        subMatchers=[
            # SM(
            #     r"\s*(?P<x_lammps_data_dihedral_coeff_list_store>{0}\s{0}\s{0}\s{0}\s{0})".format(sformat.int),
            #     repeats=True
            # )
        ])
    pair_ljcoeffs = SM(
        name='lammps-data-list-of-bonds',
        startReStr=r"\s*Pair LJCoeffs",
        repeats=True,
        subMatchers=[
            # SM(
            #     r"\s*(?P<x_lammps_data_dihedral_coeff_list_store>{0}\s{0}\s{0}\s{0}\s{0})".format(sformat.int),
            #     repeats=True
            # )
        ])
    bond_coeffs = SM(
        name='lammps-data-list-of-bonds',
        startReStr=r"\s*Bond Coeffs",
        repeats=True,
        subMatchers=[
            # SM(
            #     r"\s*(?P<x_lammps_data_dihedral_coeff_list_store>{0}\s{0}\s{0}\s{0}\s{0})".format(sformat.int),
            #     repeats=True
            # )
        ])
    angle_coeffs = SM(
        name='lammps-data-list-of-bonds',
        startReStr=r"\s*Angle Coeffs",
        repeats=True,
        subMatchers=[
            # SM(
            #     r"\s*(?P<x_lammps_data_dihedral_coeff_list_store>{0}\s{0}\s{0}\s{0}\s{0})".format(sformat.int),
            #     repeats=True
            # )
        ])
    dihedral_coeffs = SM(
        name='lammps-data-list-of-bonds',
        startReStr=r"\s*Dihedral Coeffs",
        repeats=True,
        subMatchers=[
            SM(
                r"\s*(?P<x_lammps_data_dihedral_coeff_list_store>{0}\s{0}\s{0}\s{0}\s{0})".format(sformat.int),
                repeats=True
            )
        ])

    improper_coeffs = SM(
        name='lammps-data-list-of-bonds',
        startReStr=r"\s*Improper Coeffs",
        repeats=True,
        subMatchers=[
            # SM(
            #     r"\s*(?P<x_lammps_data_dihedral_coeff_list_store>{0}\s{0}\s{0}\s{0}\s{0})".format(sformat.int),
            #     repeats=True
            # )
        ])
    bondbond_coeffs = SM(
        name='lammps-data-list-of-bonds',
        startReStr=r"\s*BondBond Coeffs",
        repeats=True,
        subMatchers=[
            # SM(
            #     r"\s*(?P<x_lammps_data_dihedral_coeff_list_store>{0}\s{0}\s{0}\s{0}\s{0})".format(sformat.int),
            #     repeats=True
            # )
        ])
    bondangle_coeffs = SM(
        name='lammps-data-list-of-bonds',
        startReStr=r"\s*BondAngle Coeffs",
        repeats=True,
        subMatchers=[
            # SM(
            #     r"\s*(?P<x_lammps_data_dihedral_coeff_list_store>{0}\s{0}\s{0}\s{0}\s{0})".format(sformat.int),
            #     repeats=True
            # )
        ])
    middlebondtorsion_coeffs = SM(
        name='lammps-data-list-of-bonds',
        startReStr=r"\s*MiddleBondTorsion Coeffs",
        repeats=True,
        subMatchers=[
            # SM(
            #     r"\s*(?P<x_lammps_data_dihedral_coeff_list_store>{0}\s{0}\s{0}\s{0}\s{0})".format(sformat.int),
            #     repeats=True
            # )
        ])
    endbondtorsion_coeffs = SM(
        name='lammps-data-list-of-bonds',
        startReStr=r"\s*EndBondTorsion Coeffs",
        repeats=True,
        subMatchers=[
            # SM(
            #     r"\s*(?P<x_lammps_data_dihedral_coeff_list_store>{0}\s{0}\s{0}\s{0}\s{0})".format(sformat.int),
            #     repeats=True
            # )
        ])
    angletorsion_coeffs = SM(
        name='lammps-data-list-of-bonds',
        startReStr=r"\s*AngleTorsion Coeffs",
        repeats=True,
        subMatchers=[
            # SM(
            #     r"\s*(?P<x_lammps_data_dihedral_coeff_list_store>{0}\s{0}\s{0}\s{0}\s{0})".format(sformat.int),
            #     repeats=True
            # )
        ])
    angleangletorsion_coeffs = SM(
        name='lammps-data-list-of-bonds',
        startReStr=r"\s*AngleAngleTorsion Coeffs",
        repeats=True,
        subMatchers=[
            # SM(
            #     r"\s*(?P<x_lammps_data_dihedral_coeff_list_store>{0}\s{0}\s{0}\s{0}\s{0})".format(sformat.int),
            #     repeats=True
            # )
        ])
    bondbond13_coeffs = SM(
        name='lammps-data-list-of-bonds',
        startReStr=r"\s*BondBond13 Coeffs",
        repeats=True,
        subMatchers=[
            # SM(
            #     r"\s*(?P<x_lammps_data_dihedral_coeff_list_store>{0}\s{0}\s{0}\s{0}\s{0})".format(sformat.int),
            #     repeats=True
            # )
        ])
    angleangle_coeffs = SM(
        name='lammps-data-list-of-bonds',
        startReStr=r"\s*AngleAngle Coeffs",
        repeats=True,
        subMatchers=[
            # SM(
            #     r"\s*(?P<x_lammps_data_dihedral_coeff_list_store>{0}\s{0}\s{0}\s{0}\s{0})".format(sformat.int),
            #     repeats=True
            # )
        ])





    return SM(
        name='root1',
        startReStr="",
        sections=['section_run'],
        forwardMatch=True,
        weak=True,
        subMatchers=[
            SM( name='root2',
                startReStr="",
                sections=['section_topology'],
                forwardMatch=True,  # The first line of the header is always skipped
                weak=True,
                # repeats=True,
                # The search is done unordered since the keywords do not appear in a specific order.
                subFlags=SM.SubFlags.Unordered,
                subMatchers=[
                    # NUMBER OF ATOMS (at_count)
                    SM(r"\s*(?P<number_of_topology_atoms>{0})\satoms".format(sformat.int), repeats=True),
                    # NUMBER OF BONDS (bd_count)
                    SM(r"\s*(?P<x_lammps_data_bd_count_store>{0})\s*bonds".format(sformat.int), repeats=True),
                    # NUMBER OF ANGLES (ag_count)
                    SM(r"\s*(?P<x_lammps_data_ag_count_store>{0})\sangles".format(sformat.int), repeats=True),
                    # NUMBER OF DIHEDRALS (dh_count)
                    SM(r"\s*(?P<x_lammps_data_dh_count_store>{0})\sdihedrals".format(sformat.int), repeats=True),
                    SM(r"\s*(?P<x_lammps_dummy>{0})\simpropers".format(sformat.int), repeats=True),
                    # NUMBER OF ATOM TYPES
                    SM(r"\s*(?P<x_lammps_data_at_types_store>{0})\satom types".format(sformat.int), repeats=True),
                    # NUMBER OF BOND TYPES
                    SM(r"\s*(?P<x_lammps_data_bd_types_store>{0})\sbond types".format(sformat.int), repeats=True),
                    # # NUMBER OF ANGLE TYPES (ag_types)
                    SM(r"\s*(?P<x_lammps_dummy>{0})\sangle types".format(sformat.int), repeats=True),
                    # # NUMBER OF DIHEDRAL TYPES (dh_types)
                    SM(r"\s*(?P<x_lammps_dummy>{0})\sdihedral types".format(sformat.int), repeats=True),
                    SM(r"\s*(?P<x_lammps_dummy>{0})\simproper types ".format(sformat.int), repeats=True),
                    SM(r"\s*(?P<x_lammps_dummy>{0})\sextra bond per atom".format(sformat.int), repeats=True),
                    SM(r"\s*(?P<x_lammps_dummy>{0})\sextra angle per atom".format(sformat.int), repeats=True),
                    SM(r"\s*(?P<x_lammps_dummy>{0})\sextra dihedral per atom".format(sformat.int), repeats=True),
                    SM(r"\s*(?P<x_lammps_dummy>{0})\sextra improper per atom ".format(sformat.int), repeats=True),
                    SM(r"\s*(?P<x_lammps_dummy>{0})\sextra special per atom ".format(sformat.int), repeats=True),
                    SM(r"\s*(?P<x_lammps_dummy>{0})\sellipsoids".format(sformat.int), repeats=True),
                    SM(r"\s*(?P<x_lammps_dummy>{0})\slines".format(sformat.int), repeats=True),
                    SM(r"\s*(?P<x_lammps_dummy>{0})\striangles".format(sformat.int), repeats=True),
                    SM(r"\s*(?P<x_lammps_dummy>{0})\sbodies".format(sformat.int), repeats=True),
                    SM(r"\s*(?P<x_lammps_dummy>{0}\s{0})\s*xlo\sxhi".format(sformat.float), repeats=True),
                    SM(r"\s*(?P<x_lammps_dummy>{0}\s{0})\s*ylo yhi".format(sformat.float), repeats=True),
                    SM(r"\s*(?P<x_lammps_dummy>{0}\s{0})\s*zlo zhi".format(sformat.float), repeats=True),
                    SM(r"\s*(?P<x_lammps_dummy>{0})\sxy xz yz".format(sformat.int), repeats=True),
                      atoms,
                    velocities,
                    ellipsoids,
                    lines,
                    triangles,
                    bodies,
                    masses,
                    bonds,
                    angles,
                    dihedrals,
                    impropers,
                    pair,
                    pair_ljcoeffs,
                    bond_coeffs,
                    angle_coeffs,
                    dihedral_coeffs,
                    improper_coeffs,
                    bondbond_coeffs,
                    bondangle_coeffs,
                    middlebondtorsion_coeffs,
                    endbondtorsion_coeffs,
                    angletorsion_coeffs,
                    angleangletorsion_coeffs,
                    bondbond13_coeffs,
                    angleangle_coeffs,

                ]
            )
        ]
    )




def get_cachingLevelForMetaName(metaInfoEnv, CachingLvl):
    """Sets the caching level for the metadata.

    Args:
        metaInfoEnv: metadata which is an object of the class InfoKindEnv in nomadcore.local_meta_info.py.
        CachingLvl: Sets the CachingLevel for the sections dos, run, and single_configuration_calculation.
            This allows to run the parser without opening new sections.

    Returns:
        Dictionary with metaname as key and caching level as value.
    """
    # manually adjust caching of metadata
    cachingLevelForMetaName = {
        'section_run': CachingLvl,
        'section_topology': CachingLvl,
    }

    # Set all temporarly metadata to Cache as they need post-processsing.
    for name in metaInfoEnv.infoKinds:
        if name.endswith('_store'):
            cachingLevelForMetaName[name] = CachingLevel.Cache
    return cachingLevelForMetaName


def main(CachingLvl):
    """Main function.

    Set up everything for the parsing of the FHI-aims DOS file and run the parsing.

    Args:
        CachingLvl: Sets the CachingLevel for the sections dos, run, and single_configuration_calculation.
            This allows to run the parser without opening new sections.
    """
    # get dos file description
    LammpsDataSimpleMatcher = build_LammpsDataFileSimpleMatcher()

    # loading metadata from nomad-meta-info/meta_info/nomad_meta_info/fhi_aims.nomadmetainfo.json
    metaInfoPath = os.path.normpath(os.path.join(os.path.dirname(os.path.abspath(__file__)), "../../../../nomad-meta-info/meta_info/nomad_meta_info/lammps.nomadmetainfo.json"))
    metaInfoEnv = get_metaInfo(metaInfoPath)

    # set parser info
    parserInfo = {'name': 'lammps-data-parser', 'version': '1.0'}

    # get caching level for metadata
    cachingLevelForMetaName = get_cachingLevelForMetaName(metaInfoEnv, CachingLvl)

    # start parsing
    mainFunction(
        mainFileDescription=LammpsDataSimpleMatcher,
        metaInfoEnv=metaInfoEnv,
        parserInfo=parserInfo,
        cachingLevelForMetaName=cachingLevelForMetaName,
        superContext=LammpsDataParserContext()
    )


if __name__ == "__main__":
    main(CachingLevel.Forward)



import numpy as np
import sys
import os
import re
import zipfile
import gzip

def get_fileExtensions(fname):
    if fname is None:
        fname = ''
    if fname != '':
        outList=[]
        if '.' in fname:
            base, ext = os.path.splitext(fname)
            if base.startswith('.'):
                base = base[1:]
            if ext.startswith('.'):
                ext = ext[1:]
            if '.' in base:
                baseList = get_fileExtensions(base)
            else:
                baseList = [base]
            if '.' in ext:
                extList = get_fileExtensions(ext)
            else:
                extList = [ext]
            for bitem in baseList:
                outList.append(bitem)
            for eitem in extList:
                outList.append(eitem)
        return outList
    else:
        return None

def get_zipType(tfile):
    basename = os.path.basename(tfile)
    exts=get_fileExtensions(basename)
    if 'gz' in exts[-1]:
        ttype = exts[-2]
        ztype = 'gz'
        zfile = basename.replace('.gz', '')
    elif 'zip' in exts[-1]:
        ttype = exts[-2]
        ztype = 'zip'
        zfile = basename.replace('.zip', '')
    else:
        ttype = exts[-1]
        ztype = None
        zfile = tfile
    return ttype, zfile, ztype

def _changes(Func):
    def newFunc(self, *args, **kwargs):
        self.need_update = True
        return Func(self, *args, **kwargs)
    return newFunc

class IndexedList(list):
    """ This list tracks the index of items 
        and assign them to its siblings _id value.
    """
    def __init__(self, *args):
        self.size = 0
        self.need_update = False
        return list.__init__(self, *args)

    def update_index(self):
        if self.need_update is True:
            for i, item in enumerate(self):
                if hasattr(item, '_id'):
                    self[i]._id=i

    @_changes
    def append(self, item):
        if hasattr(item, '_id'):
            item._id = self.size
        self.size += 1
        if hasattr(item, 'list'):
            item.list = self
        self.update_index()
        return list.append(self, item)

    @_changes
    def __delitem__(self, index):
        try:
            ind_list = range(*index.indices(len(self)))
        except AttributeError:
            ind_list = [index]

        for i in ind_list:
            if hasattr(self[i], '_id'):
                self[i]._id=-1
            if hasattr(self[i], 'list'):
                self[i].list = None

        theList = list.__delitem__(self, index)
        self.update_index()
        self.size = len(self)
        return theList

    @_changes
    def __delslice__(self, first, last):
        self.__delitem__(slice(first, last))
        self.update_index()
        self.size = len(self)

    @_changes
    def pop(self, lid=-1):
        item = list.pop(self, lid)
        if hasattr(item, '_id'):
            item._id = -1
            item.list = None
        self.update_index()
        self.size = len(self)
        return item

    @_changes
    def remove(self, item):
        try:
            list.remove(self, item)
            if hasattr(item, '_id'):
                item._id = -1
                item.list = None
            self.update_index()
            self.size = len(self)
        except AttributeError:
            pass

    @_changes
    def __setitem__(self, index, value):
        if hasattr(self[index], '_id'):
            self[index]._id=-1
        if hasattr(self[index], 'list'):
            self[index].list = None

        list.__setitem__(self, index, value)
        if hasattr(self[index], '_id'):
            self[index]._id=index
        if hasattr(self[index], 'list'):
            self[index].list = self

    #append = _changes(list.append)
    extend = _changes(list.extend)
    insert = _changes(list.insert)
    #__setitem__ = _changes(list.__setitem__)
    __iadd__ = _changes(list.__iadd__)
    __imul__ = _changes(list.__imul__)

    def __add__(self, item):
        return IndexedList(list.__add__(self, item))

    def __mul__(self, num):
        return IndexedList(list.__mul__(self, num))

class GromosTopoObject(dict):
    """ Stores data in TopoObject classes.
    """
    def __init__(self, *args, **kwargs):
        super(GromosTopoObject, self).__init__(*args, **kwargs)
        for arg in args:
            if isinstance(arg, dict):
                for k, v in arg.items():
                    if k in self:
                        self[k] = v  
        if kwargs:
            for k, v in kwargs.items():
                if k in self:
                    self[k] = v

    def __getattr__(self, attr):
        return self.get(attr)

    def __setattr__(self, key, value):
        self.__setitem__(key, value)

    def __setitem__(self, key, value):
        super(GromosTopoObject, self).__setitem__(key, value)
        self.__dict__.update({key: value})

class GromosAtomType(GromosTopoObject):

    def __init__(self, list=None, _id=-1, name='', _type='', other_type='', 
                 atom_number=0, atomic_number=-1, type_id=None, mass=0.0, charge=None, 
                 bfactor=0.0, altloc='', occupancy=0.0, solute_atom=False, 
                 solvent_atom=False, residue='', res_id=None, charge_group=None, 
                 exclude_atoms=None, lj14_atoms=None):
        self.list = list
        self._id = _id
        self.name = name.strip()
        self._type = _type.strip()
        self.type_id = type_id
        self.other_type = other_type.strip()
        self.atom_number = atom_number
        self.atomic_number = atomic_number
        self.mass = mass 
        self.charge = charge 
        self.bfactor = bfactor 
        self.altloc = altloc 
        self.occupancy = occupancy
        self.solute_atom = solute_atom 
        self.solvent_atom = solvent_atom 
        self.residue = residue
        self.res_id = res_id
        self.charge_group = charge_group
        self.exclude_atoms = exclude_atoms
        self.lj14_atoms = lj14_atoms
        self.nonbonded_list = None
        self.unique_name = ''
        if self.name != '' and self.name is not None:
            self.unique_name = self.unique_name + self.name
        if self._type != '' and self._type is not None:
            self.unique_name = self.unique_name + '_' + self._type
        if self.res_id is not None and self.residue != '':
            self.unique_name = self.unique_name + '_' + str(self.residue) + str(self.res_id)
        if self.other_type != '' and self.other_type is not None:
            self.unique_name = self.unique_name + '_' + self.other_type
        if self.unique_name == '':
            self.unique_name = None

    def __repr__(self):
        repname = '<AtomType:%s id:%d' % (self.name, self._id)
        if self.residue is not None:
            repname = repname + ', Residue:%s' % (self.residue)
        return repname + '>'

class GromosAtom(GromosTopoObject):

    def __init__(self, list=None, _id=-1, name='', atom_type=None, 
                 atom_number=None, mass=None, charge=None, bfactor=None, 
                 altloc=None, occupancy=None, solute_atom=None, 
                 solvent_atom=None, residue=None, res_id=None, charge_group=None, 
                 exclude_atoms=None, lj14_atoms=None, solvent_mol=None, 
                 mol_id=None):
        self.list = list
        self._id = _id
        self.atom_type = atom_type
        if name:
            self.name = name.strip()
        self.unique_name = ''
        if self.name != '' and self.name is not None:
            self.unique_name = self.unique_name + self.name
        if self.atom_type.name != '' and self.atom_type is not None:
            self.unique_name = self.unique_name + '_' + self.atom_type.name
        if self.atom_type._type != '' and self.atom_type is not None:
            self.unique_name = self.unique_name + '_' + self.atom_type._type
        if self.unique_name == '':
            self.unique_name = None
        if self.atom_type is not None:
            if self.mass is None: 
                self.mass = self.atom_type.mass
            if self.charge is None:
                self.charge = self.atom_type.charge
            if self.bfactor is None:
                self.bfactor = self.atom_type.bfactor
            if self.altloc is None: 
                self.altloc = self.atom_type.altloc
            if self.occupancy is None: 
                self.occupancy = self.atom_type.occupancy
            if self.solute_atom is None: 
                self.solute_atom = self.atom_type.solute_atom 
            if self.solvent_atom is None: 
                self.solvent_atom = self.atom_type.solvent_atom
            if self.residue is None: 
                self.residue = self.atom_type.residue
            if self.res_id is None: 
                self.res_id = self.atom_type.res_id
            if self.charge_group is None: 
                self.charge_group = self.atom_type.charge_group
            if self.exclude_atoms is None: 
                self.exclude_atoms = self.atom_type.exclude_atoms
            if self.lj14_atoms is None: 
                self.lj14_atoms = self.atom_type.lj14_atoms
            if self.solvent_mol is None: 
                self.solvent_mol = self.atom_type.solvent_mol

    def __repr__(self):
        repname = '<Atom:%d, %s' % (self._id, self.name)
        if self.atom_type is not None:
            repname = repname + ', Type: %s' % (self.atom_type._type)
        if self.residue:
            repname = repname + ', Residue: %s' % (self.residue)
        return repname + '>'

class GromosBondParam(GromosTopoObject):

    def __init__(self, list=None, _id=-1, name='', 
                 quartic_force=None, harmonic_force=None,
                 eq_length=None, solvent_constraint=False):
        self.list = list
        self.atoms_list = {}
        self._id = _id
        self.name = name.strip()
        self.quartic_force = quartic_force
        self.harmonic_force = harmonic_force
        self.eq_length = eq_length
        self.solvent_constraint = solvent_constraint
        self.used = False

    def __repr__(self):
        repname = '<BondParam:%d' % (self._id)
        if self.solvent_constraint:
            repname = repname + ' Solvent Constraint'
        if self.quartic_force is not None:
            repname = repname + ', CB:%f' % (self.quartic_force)
        if self.harmonic_force is not None:
            repname = repname + ', CHB:%f' % (self.harmonic_force)
        if self.eq_length is not None:
            repname = repname + ', B0:%f' % (self.eq_length)
        return repname + '>'

class GromosBondType(GromosTopoObject):

    def __init__(self, list=None, _id=-1, atom1=None, atom2=None, 
                 bond_param=None, involve_h=False):
        self.list = list
        self._id = _id
        self.atom1 = atom1
        self.atom2 = atom2
        self.atoms_list = {}
        self.bond_param = bond_param
        self.involve_h = involve_h
        self.constraint = False
        if self.bond_param is not None:
            if self.bond_param.solvent_constraint is True:
                self.constraint = True
        if self.bond_param.atoms_list is not None:
            bname = self.atom1.unique_name + ',' + self.atom2.unique_name
            if bname in self.bond_param.atoms_list:
                bname_list = self.bond_param.atoms_list[bname]
                inlist = False
                for bitem in bname_list:
                    if((self.atom1._id == bitem[0] and
                        self.atom2._id == bitem[1]) or
                       (self.atom1._id == bitem[1] and
                        self.atom2._id == bitem[0])):
                           inlist = True
                if inlist is False:
                    bname_list.append([self.atom1._id, self.atom2._id])
                    self.bond_param.atoms_list.update({bname:bname_list})
            else:
                self.bond_param.atoms_list.update({bname:[
                    [self.atom1._id, self.atom2._id]
                    ]})

    def __repr__(self):
        repname = '<BondType:%d' % (self._id)
        if self.constraint:
            repname = repname + ' Solvent Constraint'
        repname = repname + ', Param:%d AtomTypes[%d-%d]' % (
                self.bond_param._id,
                self.atom1._id, self.atom2._id)
        if self.involve_h:
            repname = repname + ' H-Bond'
        return repname + '>'

class GromosBond(GromosTopoObject):

    def __init__(self, list=None, _id=-1, atom1=None, atom2=None, 
                 bond_type=None, involve_h=False):
        self.list = list
        self._id = _id
        self.atom1 = atom1
        self.atom2 = atom2
        self.bond_type = bond_type
        self.involve_h = involve_h
        self.constraint = False
        if self.bond_type is not None:
            if self.bond_type.constraint is True:
                self.constraint = True
        if self.bond_type.atoms_list is not None:
            bname = self.atom1.unique_name + ',' + self.atom2.unique_name
            if bname in self.bond_type.atoms_list:
                bname_list = self.bond_type.atoms_list[bname]
                inlist = False
                for bitem in bname_list:
                    if((self.atom1._id == bitem[0] and
                        self.atom2._id == bitem[1]) or
                       (self.atom1._id == bitem[1] and
                        self.atom2._id == bitem[0])):
                           inlist = True
                if inlist is False:
                    bname_list.append([self.atom1._id, self.atom2._id])
                    self.bond_type.atoms_list.update({bname:bname_list})
            else:
                self.bond_type.atoms_list.update({bname:[
                    [self.atom1._id, self.atom2._id]
                    ]})

    def __repr__(self):
        repname = '<Bond:%d' % (self._id)
        if self.constraint:
            repname = repname + ' Solvent Constraint'
        repname = repname + ', Type:%d Atoms[%d-%d]' % (
                self.bond_type._id,
                self.atom1._id, self.atom2._id)
        return repname + '>'

class GromosNonbondedParam(GromosTopoObject):

    def __init__(self, list=None, _id=-1, name='', 
                 type1=None, type2=None,
                 type1_name='', type2_name='',
                 lj_r12=None, lj_r6=None,
                 lj14_r12=None, lj14_r6=None):
        self.list = list
        self._id = _id
        self.name = name.strip()
        self.type1 = type1
        self.type2 = type2
        self.type1_name = type1_name
        self.type2_name = type2_name
        #if self.type1 is not None:
        #    self.type1_name = self._atomtypes[self.type1-1]
        #if self.type2 is not None:
        #    self.type2_name = self._atomtypes[self.type2-1]
        self.unique_name = ''
        if self.type1_name != '' and self.type1 is not None:
            self.unique_name = self.type1_name
        if self.type2_name != '' and self.type2 is not None:
            self.unique_name = self.unique_name + '_' + self.type2_name
        self.lj_r12 = lj_r12
        self.lj_r6 = lj_r6
        self.lj14_r12 = lj14_r12
        self.lj14_r6 = lj14_r6

    def __repr__(self):
        repname = '<NonbondedParam:%d' % (self._id)
        if self.unique_name is not None or self.unique_name != '':
            repname = repname + ', [%s]' % (self.unique_name)
        if self.lj_r12 is not None:
            repname = repname + ', C12:%f' % (self.lj_r12)
        if self.lj_r6 is not None:
            repname = repname + ', C6:%f' % (self.lj_r6)
        if self.lj14_r12 is not None:
            repname = repname + ', CS12:%f' % (self.lj14_r12)
        if self.lj14_r6 is not None:
            repname = repname + ', CS6:%f' % (self.lj14_r6)
        return repname + '>'

class GromosNonbondedType(GromosTopoObject):

    def __init__(self, list=None, _id=-1, atom1=None, atom2=None, 
                 nonbond_param=None, solvent_mol=False):
        self.list = list
        self._id = _id
        self.atom1 = atom1
        self.atom2 = atom2
        self.nonbond_param = nonbond_param
        self.involve_h = involve_h

    def __repr__(self):
        repname = '<NonbondedType:%d Param:%d AtomTypes[%d-%d]' % (
                self._id, self.nonbond_param._id,
                self.atom1._id, self.atom2._id)
        return repname + '>'


class GromosAngleParam(GromosTopoObject):

    def __init__(self, list=None, _id=-1, name='', 
                 harmonic_cosine_force=None, harmonic_force=None,
                 eq_angle=None):
        self.list = list
        self.atoms_list = {}
        self._id = _id
        self.name = name.strip()
        self.harmonic_cosine_force = harmonic_cosine_force
        self.harmonic_force = harmonic_force
        self.eq_angle = eq_angle
        self.used = False

    def __repr__(self):
        repname = '<AngleParam:%d' % (self._id)
        if self.harmonic_cosine_force is not None:
            repname = repname + ', CT:%f' % (self.harmonic_cosine_force)
        if self.harmonic_force is not None:
            repname = repname + ', CHT:%f' % (self.harmonic_force)
        if self.eq_angle is not None:
            repname = repname + ', T0:%f' % (self.eq_angle)
        return repname + '>'

class GromosAngleType(GromosTopoObject):

    def __init__(self, list=None, _id=-1, atom1=None, atom2=None, 
                 atom3=None, angle_param=None, involve_h=False):
        self.list = list
        self._id = _id
        self.atom1 = atom1
        self.atom2 = atom2
        self.atom3 = atom3
        self.angle_param = angle_param
        self.involve_h = involve_h
        angname = self.atom1.unique_name 
        angname = angname + ',' + self.atom2.unique_name 
        angname = angname + ',' + self.atom3.unique_name 
        angname_list = []
        if self.angle_param.atoms_list is not None:
            if angname in self.angle_param.atoms_list:
                angname_list = self.angle_param.atoms_list[angname]
        inlist = False
        for angitem in angname_list:
            if((self.atom1._id == angitem[0] and
                self.atom2._id == angitem[1] and
                self.atom3._id == angitem[2]) or
               (self.atom1._id == angitem[2] and
                self.atom2._id == angitem[1] and
                self.atom3._id == angitem[0])):
                   inlist = True
        if inlist is False:
            angname_list.append([
                self.atom1._id, 
                self.atom2._id,
                self.atom3._id
                ])
            self.angle_param.atoms_list.update(
                    {angname:angname_list})
        else:
            self.angle_param.atoms_list.update({angname:[
                [self.atom1._id, 
                 self.atom2._id,
                 self.atom3._id]
                ]})

    def __repr__(self):
        repname = '<AngleType:%d Param:%d AtomTypes[%d-%d-%d]' % (
                self._id, self.angle_param._id, self.atom1._id, 
                self.atom2._id, self.atom3._id)
        return repname + '>'

class GromosDihedralParam(GromosTopoObject):

    def __init__(self, list=None, _id=-1, name='', improper=False,
                 force_const=None, eq_phase_angle=None,
                 multiplicity=None):
        self.list = list
        self.dihedral_atoms_list = {}
        self.improper_atoms_list = {}
        self._id = _id
        self.name = name.strip()
        self.improper = improper
        self.force_const = force_const
        self.eq_phase_angle = eq_phase_angle
        self.multiplicity = multiplicity
        self.used = False

    def __repr__(self):
        repname = '<DihedralParam:%d' % (self._id)
        if self.improper is True:
            repname = repname + ' Improper'
        if self.force_const is not None:
            repname = repname + ', C:%f' % (self.force_const)
        if self.eq_phase_angle is not None:
            repname = repname + ', Q0/PD:%f' % (self.eq_phase_angle)
        return repname + '>'

class GromosDihedralType(GromosTopoObject):

    def __init__(self, list=None, _id=-1, atom1=None, atom2=None, 
                 atom3=None, atom4=None, dihedral_param=None, 
                 involve_h=False):
        self.list = list
        self._id = _id
        self.atom1 = atom1
        self.atom2 = atom2
        self.atom3 = atom3
        self.atom4 = atom4
        if dihedral_param.improper:
            self.improper = True
        else:
            self.improper = False
        self.dihedral_param = dihedral_param
        self.involve_h = involve_h
        dname = self.atom1.unique_name 
        dname = dname + ',' + self.atom2.unique_name 
        dname = dname + ',' + self.atom3.unique_name 
        dname = dname + ',' + self.atom4.unique_name 
        dname_list=[]
        if self.improper:
            if self.dihedral_param.improper_atoms_list is not None:
                if dname in self.dihedral_param.improper_atoms_list:
                    dname_list = self.dihedral_param.improper_atoms_list[dname]
        else:
            if self.dihedral_param.dihedral_atoms_list is not None:
                if dname in self.dihedral_param.dihedral_atoms_list:
                    dname_list = self.dihedral_param.dihedral_atoms_list[dname]
        inlist = False
        for ditem in dname_list:
            if((self.atom1._id == ditem[0] and
                self.atom2._id == ditem[1] and
                self.atom3._id == ditem[2] and
                self.atom4._id == ditem[3]) or
               (self.atom1._id == ditem[3] and
                self.atom2._id == ditem[2] and
                self.atom3._id == ditem[1] and
                self.atom4._id == ditem[0])):
                   inlist = True
        if inlist is False:
            dname_list.append([
                self.atom1._id, 
                self.atom2._id,
                self.atom3._id,
                self.atom4._id
                ])
            if self.improper:
                self.dihedral_param.improper_atoms_list.update(
                        {dname:dname_list})
            else:
                self.dihedral_param.dihedral_atoms_list.update(
                        {dname:dname_list})
        else:
            if self.improper:
                self.dihedral_param.improper_atoms_list.update({dname:[
                    [self.atom1._id, 
                     self.atom2._id,
                     self.atom3._id,
                     self.atom4._id]
                    ]})
            else:
                self.dihedral_param.dihedral_atoms_list.update({dname:[
                    [self.atom1._id, 
                     self.atom2._id,
                     self.atom3._id,
                     self.atom4._id]
                    ]})

    def __repr__(self):
        repname = '<DihedralType:%d' % (self._id) 
        if self.improper is True:
            repname = repname + ' Improper'
        if self.dihedral_param is not None:
            repname = repname + ' Param:%d' % (self.dihedral_param._id)
        if self.improper_param is not None:
            repname = repname + ' Param:%d' % (self.improper_param._id)
        repname = repname + ' AtomTypes[%d-%d-%d-%d]' % (self.atom1._id, 
                self.atom2._id, self.atom3._id, self.atom4._id)
        return repname + '>'

class GromosTopology(GromosTopoObject):

    def __init__(self, top=None, cnf=None,
                 top_format=None, cnf_format=None): 
        self.reset_top()
        self.reset_cnf()
        self.load(topofile=top, topotype=top_format,
                  conffile=cnf, conftype=cnf_format)

    def reset_top(self):
        self.atomtypes = None
        self._atomtypes = None
        self.bondtypes = None
        self.bondparams = None
        self._bondparams = None
        self.angletypes = None
        self._angleparams = None
        self.angleparams = None
        self._dihedralparams = None
        self.dihedralparams = None
        self.improperparams = None
        self._improper_param_ids = None
        self._dihedral_param_ids = None
        self.impropertypes = None
        self.dihedraltypes = None
        self.crossterms = None
        self._ljparameters = None
        self.ljparameters = None
        self.exception_nonbonds = None
        self.residues = None
        self._residues = None
        self.solutemols = None
        self.molecules = None
        self.pressure_groups = None
        self.temperature_groups = None
        self.solventatoms = None
        self.solvent_constraints = None
        self.solvent = False
        self.physical_constants = None
        self.version = None
        self._id = None
        self.title = None

    def reset_cnf(self):
        self.conf_title = None
        self.nbounds = None
        self.time = None
        self.step = None
        self.atoms = None
        self.bonds = None
        self.angles = None
        self.dihedrals = None
        self.impropers = None
        self.positions = None
        self.velocities = None
        self.a = None # box dimensions
        self.b = None
        self.c = None
        self.alpha = None # box angles
        self.beta = None
        self.gamma = None
        self.phi = None # rotation angles
        self.theta= None
        self.psi = None
        self.X = None
        self.Y = None
        self.Z = None

    def load(self, topofile=None, topotype=None,
             conffile=None, conftype=None, 
             clear_top=False, clear_cnf=True):
        if topofile is not None:
            self.load_top(topofile, topotype, clear=clear_top)
        if conffile is not None:
            self.load_cnf(conffile, conftype, clear=clear_cnf)

    def load_top(self, topofile, topotype, clear=False):
        if clear:
            self.reset_top()
        readOK=False
        ztype=None
        zfile=None
        if topotype is None:
            topotype, zfile, ztype = get_zipType(topofile)
        if('gromostop' in topotype or
           'top' in topotype):
            typtop, zfile, ztype = get_zipType(topofile)
            if topofile is not None:
                readOK=True
        if readOK:
            if ztype is not None:
                if 'zip' in ztype:
                    with zipfile.ZipFile(topofile) as zin:
                        with zin.open(zfile, 'r') as fin:
                            self.read_top(fin, decode=True)
                elif 'gz' in ztype:
                    with gzip.open(topofile, 'rt') as fin:
                        self.read_top(fin)
            else:
                with open(topofile, 'r') as fin:
                    self.read_top(fin)
        if self._bondparams:
            if self.bondparams is None:
                self.bondparams = []
            for bparam in self._bondparams:
                if bparam.atoms_list:
                    self.bondparams.append(bparam)
        if self._angleparams:
            if self.angleparams is None:
                self.angleparams = []
            for angparam in self._angleparams:
                if angparam.atoms_list:
                    self.angleparams.append(angparam)
        if self._dihedralparams:
            if self.improperparams is None:
                self.improperparams = []
            if self.dihedralparams is None:
                self.dihedralparams = []
            for dparam in self._dihedralparams:
                if dparam.improper:
                    if dparam.improper_atoms_list:
                        self.improperparams.append(dparam)
                else:
                    if dparam.dihedral_atoms_list:
                        self.dihedralparams.append(dparam)
        if self._ljparameters:
            if self.ljparameters is None:
                self.ljparameters = []
            typenames=[atype._type for atype in self.atomtypes]
            for pari, parm in enumerate(self._ljparameters):
                if(parm.type1_name in typenames and
                   parm.type2_name in typenames):
                    self.ljparameters.append(self._ljparameters[pari])

    def load_cnf(self, conffile, conftype, clear=True):
        if clear:
            self.reset_cnf()
        readOK=False
        ztype=None
        zfile=None
        if conftype is None:
            conftype, zfile, ztype = get_zipType(conffile)
        if('gromoscnf' in conftype or
           'gromos96' in conftype or
           'cnf' in conftype or
           'g96' in conftype):
            typcnf, zfile, ztype = get_zipType(conffile)
            if conffile is not None:
                readOK=True
        if readOK:
            if ztype is not None:
                if 'zip' in ztype:
                    with zipfile.ZipFile(conffile) as zin:
                        with zin.open(zfile, 'r') as fin:
                            self.read_cnf(fin, decode=True)
                elif 'gz' in ztype:
                    with gzip.open(conffile, 'rt') as fin:
                        self.read_cnf(fin)
            else:
                with open(conffile, 'r') as fin:
                    self.read_cnf(fin)
            if self._residues is not None:
                if self.residues is None:
                    self.residues = {}
                for res in self._residues:
                    reslist=[]
                    for ai, at in enumerate(self.atoms):
                        if res in at.residue:
                            reslist.append(self.atoms[ai])
                    self.residues.update({res:reslist})

    def read_top(self, fin, decode=False):
        emptyLine = re.compile(r"^\s*$")
        physc=0
        atn=0
        atnum=0
        atset=False
        resn=0
        resnum=0
        resset=False
        satn=0
        satnum=0
        satnumset=False
        satset=False
        satnn=0
        solvn=0
        solvnum=0
        solvnumset=False
        solvset=False
        solvnn=0
        atnm = None
        mres = None
        panm = None
        iac = None
        mass = None
        cg = None
        cgc = None
        ine = -1
        inelist = []
        ine14 = -1
        ine14list = []
        btn=0
        btnum=0
        btnumset=False
        btset=False
        btnn=0
        cb = None
        cbh = None
        b0 = None
        #--
        bhn=0
        bhnum=0
        bhnumset=False
        bhset=False
        bhnn=0
        ibh = None
        jbh = None
        icbh = None
        #--
        bn=0
        bnum=0
        bnumset=False
        bset=False
        bnn=0
        ib = None
        jb = None
        ic = None
        jc = None
        icb = None
        #--
        angtn=0
        angtnum=0
        angtnumset=False
        angtset=False
        angtnn=0
        ct = None
        cth = None
        t0 = None
        #--
        anghn=0
        anghnum=0
        anghnumset=False
        anghset=False
        anghnn=0
        ith = None
        jth = None
        kth = None
        icth = None
        #--
        angn=0
        angnum=0
        angnumset=False
        angset=False
        angnn=0
        it = None
        jt = None
        kt = None
        ict = None
        #--
        imptn=0
        imptnum=0
        imptnumset=False
        imptset=False
        imptnn=0
        dihtn=0
        dihtnum=0
        dihtnumset=False
        dihtset=False
        dihtnn=0
        cq = None
        q0 = None
        np = None
        #--
        imphn=0
        imphnum=0
        imphnumset=False
        imphset=False
        imphnn=0
        dihhn=0
        dihhnum=0
        dihhnumset=False
        dihhset=False
        dihhnn=0
        iqh = None
        jqh = None
        kqh = None
        lqh = None
        icqh = None
        #--
        impn=0
        impnum=0
        impnumset=False
        impset=False
        impnn=0
        dihn=0
        dihnum=0
        dihnumset=False
        dihset=False
        dihnn=0
        iq = None
        jq = None
        kq = None
        lq = None
        icq = None
        #--
        ljn=0
        ljnum=0
        ljnumset=False
        ljset=False
        ljnn=0
        ljen=0
        ljenum=0
        ljenumset=False
        ljeset=False
        ljenn=0
        ilj = None
        jlj = None
        c12 = None
        c6 = None
        cs12 = None
        cs6 = None
        #--
        smn=0
        smnum=0
        smnumset=False
        smset=False
        smnn=0
        #--
        scn=0
        scnum=0
        scnumset=False
        scset=False
        scnn=0
        #--
        tgn=0
        tgnum=0
        tgnumset=False
        tgset=False
        tgnn=0
        #--
        pgn=0
        pgnum=0
        pgnumset=False
        pgset=False
        pgnn=0
        for line in fin:
            if decode:
                line = line.decode('utf-8')
            if emptyLine.findall(line):
                continue
            cmdLine = ' '.join([x for x in line.strip().split() if x])
            cmdLineUp = cmdLine.upper()
            if cmdLineUp.startswith('#'):
                continue
            if cmdLineUp.startswith('END'):
                section = None
            elif cmdLineUp.startswith('TITLE'):
                section = 'TITLE'
                continue
            elif cmdLineUp.startswith('PHYSICALCONSTANTS'):
                section = 'PHYSICALCONSTANTS'
                continue
            elif cmdLineUp.startswith('TOPVERSION'):
                section = 'TOPVERSION'
                continue
            elif cmdLineUp.startswith('ATOMTYPENAME'):
                section = 'ATOMTYPENAME'
                continue
            elif cmdLineUp.startswith('RESNAME'):
                section = 'RESNAME'
                continue
            elif cmdLineUp.startswith('SOLUTEATOM'):
                section = 'SOLUTEATOM'
                continue
            elif cmdLineUp.startswith('BONDSTRETCHTYPE'):
                section = 'BONDSTRETCHTYPE'
                continue
            elif cmdLineUp.startswith('BONDH'):
                section = 'BONDH'
                continue
            elif cmdLineUp.split()[0] == 'BOND':
                section = 'BOND'
                continue
            elif cmdLineUp.startswith('BONDANGLEBENDTYPE'):
                section = 'BONDANGLEBENDTYPE'
                continue
            elif cmdLineUp.startswith('BONDANGLEH'):
                section = 'BONDANGLEH'
                continue
            elif cmdLineUp.split()[0] == 'BONDANGLE':
                section = 'BONDANGLE'
                continue
            elif cmdLineUp.startswith('IMPDIHEDRALTYPE'):
                section = 'IMPDIHEDRALTYPE'
                continue
            elif cmdLineUp.startswith('IMPDIHEDRALH'):
                section = 'IMPDIHEDRALH'
                continue
            elif cmdLineUp.split()[0] == 'IMPDIHEDRAL':
                section = 'IMPDIHEDRAL'
                continue
            elif cmdLineUp.startswith('TORSDIHEDRALTYPE'):
                section = 'TORSDIHEDRALTYPE'
                continue
            elif cmdLineUp.startswith('DIHEDRALH'):
                section = 'DIHEDRALH'
                continue
            elif cmdLineUp.split()[0] == 'DIHEDRAL':
                section = 'DIHEDRAL'
                continue
            elif cmdLineUp.startswith('LJPARAMETERS'):
                section = 'LJPARAMETERS'
                continue
            elif cmdLineUp.startswith('SOLUTEMOLECULES'):
                section = 'SOLUTEMOLECULES'
                continue
            elif cmdLineUp.startswith('TEMPERATUREGROUPS'):
                section = 'TEMPERATUREGROUPS'
                continue
            elif cmdLineUp.startswith('PRESSUREGROUPS'):
                section = 'PRESSUREGROUPS'
                continue
            elif cmdLineUp.startswith('LJEXCEPTIONS'):
                section = 'LJEXCEPTIONS'
                continue
            elif cmdLineUp.startswith('SOLVENTATOM'):
                section = 'SOLVENTATOM'
                continue
            elif cmdLineUp.startswith('SOLVENTCONSTR'):
                section = 'SOLVENTCONSTR'
                continue
            if section == 'TITLE':
                if self.title is None:
                    self.title = ''
                self.title = self.title + ' ' + cmdLine
            if section == 'PHYSICALCONSTANTS':
                pok=False
                if self.physical_constants is None:
                    self.physical_constants = {}
                val = cmdLine.split()[0]
                try:
                    val = float(val)
                except(TypeError, ValueError):
                    pass
                if physc == 0:
                    pok=True
                    self.physical_constants.update({'FPEPSI' : val})
                elif physc == 1:
                    pok=True
                    self.physical_constants.update({'HBAR' : val})
                elif physc == 2:
                    pok=True
                    self.physical_constants.update({'SPDL-nm_ps' : val})
                elif physc == 3:
                    pok=True
                    self.physical_constants.update({'BOLTZ' : val})
                if pok:
                    physc += 1
            if section == 'TOPVERSION':
                self.version = cmdLine.split()[0]
            if section == 'ATOMTYPENAME':
                atok=False
                val = cmdLine.split()[0]
                if atn == 0:
                    atok=True
                    try:
                        atnum = int(val)
                        atset = True
                    except(TypeError, ValueError):
                        atset = False
                else:
                    atok=True
                    if self._atomtypes is None:
                        self._atomtypes = []
                    self._atomtypes.append(val)
                    if atset is False:
                        atnum = len(self._atomtypes)
                if atok:
                    atn += 1
            if section == 'RESNAME':
                resok=False
                val = cmdLine.split()[0]
                if resn == 0:
                    resok=True
                    try:
                        resnum = int(val)
                        resset = True
                    except(TypeError, ValueError):
                        resset = False
                else:
                    resok=True
                    if self._residues is None:
                        self._residues = []
                    self._residues.append(val)
                    if resset is False:
                        resnum = len(self._residues)
                if resok:
                    resn += 1
            if section == 'SOLUTEATOM':
                satok=False
                if satn == 0:
                    satok=True
                    try:
                        val = cmdLine.split()[0]
                        satnum = int(val)
                        satnumset = True
                    except(TypeError, ValueError):
                        satnumset = False
                else:
                    if satset is False:
                        for item in cmdLine.split():
                            if satnn == 0:
                                atnm = int(item)
                            elif satnn == 1:
                                mres = int(item)
                            elif satnn == 2:
                                panm = item
                            elif satnn == 3:
                                iac = int(item)
                            elif satnn == 4:
                                mass = float(item)
                            elif satnn == 5:
                                cg = float(item)
                            elif satnn == 6:
                                cgc = int(item)
                            elif satnn == 7:
                                ine = int(item)
                            elif satnn>7 and ine>0:
                                if satnn>7 and satnn<=7+ine:
                                    inelist.append(int(item))
                                elif satnn>7+ine and ine14<0:
                                    ine14 = int(item)
                                    if ine14<1:
                                        satset=True
                                elif((satnn>8+ine and satnn<=8+ine+ine14) and 
                                      ine14>0):
                                    ine14list.append(int(item))
                                    satset = True
                            elif satnn>7 and ine<1 and ine14<0:
                                ine14 = int(item)
                                if ine14<1:
                                    satset=True
                            elif((satnn>8 and satnn<=8+ine14) and 
                                  ine<1 and ine14>0):
                                ine14list.append(int(item))
                                satset = True
                            satnn += 1
                    if satset is True:
                        if self.atomtypes is None:
                            self.atomtypes = IndexedList()
                        self.atomtypes.append(GromosAtomType(name=panm, 
                            _type=self._atomtypes[iac-1], type_id=iac, 
                            atom_number=atnm, mass=mass, charge=cg, 
                            solute_atom=True, charge_group=cgc, 
                            res_id=mres, residue=self._residues[mres-1], 
                            exclude_atoms=inelist, lj14_atoms=ine14list))
                        satnn=0
                        atnm = None
                        mres = None
                        panm = None
                        iac = None
                        mass = None
                        cg = None
                        cgc = None
                        ine = -1
                        inelist = []
                        ine14 = -1
                        ine14list = []
                        satset = False
                    satok=True
                if satok:
                    satn += 1
            if section == 'BONDSTRETCHTYPE':
                btok=False
                if btn == 0:
                    btok=True
                    try:
                        val = cmdLine.split()[0]
                        btnum = int(val)
                        btnumset = True
                    except(TypeError, ValueError):
                        btnumset = False
                else:
                    if btset is False:
                        for item in cmdLine.split():
                            if btnn < 1:
                                cb = float(item)
                            elif btnn < 2:
                                chb = float(item)
                            elif btnn < 3:
                                b0 = float(item)
                                btset = True
                            btnn += 1
                    if btset is True:
                        if self._bondparams is None:
                            self._bondparams = IndexedList()
                        self._bondparams.append(GromosBondParam(
                            quartic_force=cb, harmonic_force=chb, 
                            eq_length=b0))
                        btnn=0
                        cb = None
                        cbh = None
                        b0 = None
                        btset = False
                    btok=True
                if btok:
                    btn += 1
            if section == 'BONDH':
                bhok=False
                if bhn == 0:
                    bhok=True
                    try:
                        val = cmdLine.split()[0]
                        bhnum = int(val)
                        bhnumset = True
                    except(TypeError, ValueError):
                        bhnumset = False
                else:
                    if bhset is False:
                        for item in cmdLine.split():
                            if bhnn < 1:
                                ibh = int(item)
                            elif bhnn < 2:
                                jbh = int(item)
                            elif bhnn < 3:
                                icbh = int(item)
                                bhset = True
                            bhnn += 1
                    if bhset is True:
                        if self.bondtypes is None:
                            self.bondtypes = IndexedList()
                        self._bondparams[icbh-1].used=True
                        self.bondtypes.append(GromosBondType(
                            atom1=self.atomtypes[ibh-1], 
                            atom2=self.atomtypes[jbh-1], 
                            bond_param=self._bondparams[icbh-1], 
                            involve_h=True
                            ))
                        bhnn=0
                        ibh = None
                        jbh = None
                        icbh = None
                        bhset = False
                    bhok=True
                if bhok:
                    bhn += 1
            if section == 'BOND':
                bok=False
                if bn == 0:
                    bok=True
                    try:
                        val = cmdLine.split()[0]
                        bnum = int(val)
                        bnumset = True
                    except(TypeError, ValueError):
                        bnumset = False
                else:
                    if bset is False:
                        for item in cmdLine.split():
                            if bnn < 1:
                                ib = int(item)
                            elif bnn < 2:
                                jb = int(item)
                            elif bnn < 3:
                                icb = int(item)
                                bset = True
                            bnn += 1
                    if bset is True:
                        if self.bondtypes is None:
                            self.bondtypes = IndexedList()
                        self._bondparams[icb-1].used=True
                        self.bondtypes.append(GromosBondType(
                            atom1=self.atomtypes[ib-1], 
                            atom2=self.atomtypes[jb-1], 
                            bond_param=self._bondparams[icb-1]
                            ))
                        bnn=0
                        ib = None
                        jb = None
                        icb = None
                        bset = False
                    bok=True
                if bok:
                    bn += 1
            if section == 'BONDANGLEBENDTYPE':
                angtok=False
                if angtn == 0:
                    angtok=True
                    try:
                        val = cmdLine.split()[0]
                        angtnum = int(val)
                        angtnumset = True
                    except(TypeError, ValueError):
                        angtnumset = False
                else:
                    if angtset is False:
                        for item in cmdLine.split():
                            if angtnn < 1:
                                ct = float(item)
                            elif angtnn < 2:
                                cht = float(item)
                            elif angtnn < 3:
                                t0 = float(item)
                                angtset = True
                            angtnn += 1
                    if angtset is True:
                        if self._angleparams is None:
                            self._angleparams = IndexedList()
                        self._angleparams.append(GromosAngleParam(
                            harmonic_cosine_force=ct, 
                            harmonic_force=cht, eq_angle=t0))
                        angtnn=0
                        ct = None
                        cht = None
                        t0 = None
                        angtset = False
                    angtok=True
                if angtok:
                    angtn += 1
            if section == 'BONDANGLEH':
                anghok=False
                if anghn == 0:
                    anghok=True
                    try:
                        val = cmdLine.split()[0]
                        anghnum = int(val)
                        anghnumset = True
                    except(TypeError, ValueError):
                        anghnumset = False
                else:
                    if anghset is False:
                        for item in cmdLine.split():
                            if anghnn < 1:
                                ith = int(item)
                            elif anghnn < 2:
                                jth = int(item)
                            elif anghnn < 3:
                                kth = int(item)
                            elif anghnn < 4:
                                icth = int(item)
                                anghset = True
                            anghnn += 1
                    if anghset is True:
                        if self.angletypes is None:
                            self.angletypes = IndexedList()
                        self._angleparams[icth-1].used=True
                        self.angletypes.append(GromosAngleType(
                            atom1=self.atomtypes[ith-1], 
                            atom2=self.atomtypes[jth-1], 
                            atom3=self.atomtypes[kth-1], 
                            angle_param=self._angleparams[icth-1],
                            involve_h=True
                            ))
                        anghnn=0
                        ith = None
                        jth = None
                        kth = None
                        icth = None
                        anghset = False
                    anghok=True
                if anghok:
                    anghn += 1
            if section == 'BONDANGLE':
                angok=False
                if angn == 0:
                    angok=True
                    try:
                        val = cmdLine.split()[0]
                        angnum = int(val)
                        angnumset = True
                    except(TypeError, ValueError):
                        angnumset = False
                else:
                    if angset is False:
                        for item in cmdLine.split():
                            if angnn < 1:
                                it = int(item)
                            elif angnn < 2:
                                jt = int(item)
                            elif angnn < 3:
                                kt = int(item)
                            elif angnn < 4:
                                ict = int(item)
                                angset = True
                            angnn += 1
                    if angset is True:
                        if self.angletypes is None:
                            self.angletypes = IndexedList()
                        self._angleparams[ict-1].used=True
                        self.angletypes.append(GromosAngleType(
                            atom1=self.atomtypes[it-1], 
                            atom2=self.atomtypes[jt-1], 
                            atom3=self.atomtypes[kt-1], 
                            angle_param=self._angleparams[ict-1]
                            ))
                        angnn=0
                        it = None
                        jt = None
                        kt = None
                        ict = None
                        angset = False
                    angok=True
                if angok:
                    angn += 1
            if section == 'IMPDIHEDRALTYPE':
                imptok=False
                if imptn == 0:
                    imptok=True
                    try:
                        val = cmdLine.split()[0]
                        imptnum = int(val)
                        imptnumset = True
                    except(TypeError, ValueError):
                        imptnumset = False
                else:
                    if imptset is False:
                        for item in cmdLine.split():
                            if imptnn < 1:
                                cq = float(item)
                            elif imptnn < 2:
                                q0 = float(item)
                                imptset = True
                            imptnn += 1
                    if imptset is True:
                        if self._dihedralparams is None:
                            self._dihedralparams = IndexedList()
                        self._dihedralparams.append(GromosDihedralParam(
                            improper=True, force_const=cq, 
                            eq_phase_angle=q0))
                        if self._improper_param_ids is None:
                            self._improper_param_ids = []
                        self._improper_param_ids.append(
                                len(self._dihedralparams)-1)
                        imptnn=0
                        cq = None
                        q1 = None
                        imptset = False
                    imptok=True
                if imptok:
                    imptn += 1
            if section == 'IMPDIHEDRALH':
                imphok=False
                if imphn == 0:
                    imphok=True
                    try:
                        val = cmdLine.split()[0]
                        imphnum = int(val)
                        imphnumset = True
                    except(TypeError, ValueError):
                        imphnumset = False
                else:
                    if imphset is False:
                        for item in cmdLine.split():
                            if imphnn < 1:
                                iqh = int(item)
                            elif imphnn < 2:
                                jqh = int(item)
                            elif imphnn < 3:
                                kqh = int(item)
                            elif imphnn < 4:
                                lqh = int(item)
                            elif imphnn < 5:
                                icqh = int(item)
                                imphset = True
                            imphnn += 1
                    if imphset is True:
                        if self.impropertypes is None:
                            self.impropertypes = IndexedList()
                        self._dihedralparams[
                                self._improper_param_ids[
                                    icqh-1]].used=True
                        self.impropertypes.append(GromosDihedralType(
                            atom1=self.atomtypes[iqh-1], 
                            atom2=self.atomtypes[jqh-1], 
                            atom3=self.atomtypes[kqh-1], 
                            atom4=self.atomtypes[lqh-1], 
                            dihedral_param=self._dihedralparams[
                                self._improper_param_ids[icqh-1]],
                            involve_h=True
                            ))
                        imphnn=0
                        iqh = None
                        jqh = None
                        kqh = None
                        lqh = None
                        icqh = None
                        imphset = False
                    imphok=True
                if imphok:
                    imphn += 1
            if section == 'IMPDIHEDRAL':
                impok=False
                if impn == 0:
                    impok=True
                    try:
                        val = cmdLine.split()[0]
                        impnum = int(val)
                        impnumset = True
                    except(TypeError, ValueError):
                        impnumset = False
                else:
                    if impset is False:
                        for item in cmdLine.split():
                            if impnn < 1:
                                iq = int(item)
                            elif impnn < 2:
                                jq = int(item)
                            elif impnn < 3:
                                kq = int(item)
                            elif impnn < 4:
                                lq = int(item)
                            elif impnn < 5:
                                icq = int(item)
                                impset = True
                            impnn += 1
                    if impset is True:
                        if self.impropertypes is None:
                            self.impropertypes = IndexedList()
                        self._dihedralparams[
                                self._improper_param_ids[
                                    icq-1]].used=True
                        self.impropertypes.append(GromosDihedralType(
                            atom1=self.atomtypes[iq-1], 
                            atom2=self.atomtypes[jq-1], 
                            atom3=self.atomtypes[kq-1], 
                            atom4=self.atomtypes[lq-1], 
                            dihedral_param=self._dihedralparams[
                                self._improper_param_ids[icq-1]]
                            ))
                        impnn=0
                        iq = None
                        jq = None
                        kq = None
                        lq = None
                        icq = None
                        impset = False
                    impok=True
                if impok:
                    impn += 1
            if section == 'TORSDIHEDRALTYPE':
                dihtok=False
                if dihtn == 0:
                    dihtok=True
                    try:
                        val = cmdLine.split()[0]
                        dihtnum = int(val)
                        dihtnumset = True
                    except(TypeError, ValueError):
                        dihtnumset = False
                else:
                    if dihtset is False:
                        for item in cmdLine.split():
                            if dihtnn < 1:
                                cq = float(item)
                            elif dihtnn < 2:
                                q0 = float(item)
                            elif dihtnn < 3:
                                np = int(item)
                                dihtset = True
                            dihtnn += 1
                    if dihtset is True:
                        if self._dihedralparams is None:
                            self._dihedralparams = IndexedList()
                        self._dihedralparams.append(GromosDihedralParam(
                            force_const=cq, eq_phase_angle=q0,
                            multiplicity=np))
                        if self._dihedral_param_ids is None:
                            self._dihedral_param_ids = []
                        self._dihedral_param_ids.append(
                                len(self._dihedralparams)-1)
                        dihtnn=0
                        cq = None
                        q1 = None
                        np = None
                        dihtset = False
                    dihtok=True
                if dihtok:
                    dihtn += 1
            if section == 'DIHEDRALH':
                dihhok=False
                if dihhn == 0:
                    dihhok=True
                    try:
                        val = cmdLine.split()[0]
                        dihhnum = int(val)
                        dihhnumset = True
                    except(TypeError, ValueError):
                        dihhnumset = False
                else:
                    if dihhset is False:
                        for item in cmdLine.split():
                            if dihhnn < 1:
                                iqh = int(item)
                            elif dihhnn < 2:
                                jqh = int(item)
                            elif dihhnn < 3:
                                kqh = int(item)
                            elif dihhnn < 4:
                                lqh = int(item)
                            elif dihhnn < 5:
                                icqh = int(item)
                                dihhset = True
                            dihhnn += 1
                    if dihhset is True:
                        if self.dihedraltypes is None:
                            self.dihedraltypes = IndexedList()
                        self._dihedralparams[
                                self._dihedral_param_ids[
                                    icqh-1]].used=True
                        self.dihedraltypes.append(GromosDihedralType(
                            atom1=self.atomtypes[iqh-1], 
                            atom2=self.atomtypes[jqh-1], 
                            atom3=self.atomtypes[kqh-1], 
                            atom4=self.atomtypes[lqh-1], 
                            dihedral_param=self._dihedralparams[
                                self._dihedral_param_ids[icqh-1]],
                            involve_h=True
                            ))
                        dihhnn=0
                        iqh = None
                        jqh = None
                        kqh = None
                        lqh = None
                        icqh = None
                        dihhset = False
                    dihhok=True
                if dihhok:
                    dihhn += 1
            if section == 'DIHEDRAL':
                dihok=False
                if dihn == 0:
                    dihok=True
                    try:
                        val = cmdLine.split()[0]
                        dihnum = int(val)
                        dihnumset = True
                    except(TypeError, ValueError):
                        dihnumset = False
                else:
                    if dihset is False:
                        for item in cmdLine.split():
                            if dihnn < 1:
                                iq = int(item)
                            elif dihnn < 2:
                                jq = int(item)
                            elif dihnn < 3:
                                kq = int(item)
                            elif dihnn < 4:
                                lq = int(item)
                            elif dihnn < 5:
                                icq = int(item)
                                dihset = True
                            dihnn += 1
                    if dihset is True:
                        if self.dihedraltypes is None:
                            self.dihedraltypes = IndexedList()
                        self._dihedralparams[
                                self._dihedral_param_ids[
                                    icq-1]].used=True
                        self.dihedraltypes.append(GromosDihedralType(
                            atom1=self.atomtypes[iq-1], 
                            atom2=self.atomtypes[jq-1], 
                            atom3=self.atomtypes[kq-1], 
                            atom4=self.atomtypes[lq-1], 
                            dihedral_param=self._dihedralparams[
                                self._dihedral_param_ids[icq-1]]
                            ))
                        dihnn=0
                        iq = None
                        jq = None
                        kq = None
                        lq = None
                        icq = None
                        dihset = False
                    dihok=True
                if dihok:
                    dihn += 1
            if section == 'LJPARAMETERS':
                ljok=False
                if ljn == 0:
                    ljok=True
                    try:
                        val = cmdLine.split()[0]
                        ljnum = int(val)
                        ljnumset = True
                    except(TypeError, ValueError):
                        ljnumset = False
                else:
                    if ljset is False:
                        for item in cmdLine.split():
                            if ljnn < 1:
                                ilj = int(item)
                            elif ljnn < 2:
                                jlj = int(item)
                            elif ljnn < 3:
                                c12 = float(item)
                            elif ljnn < 4:
                                c6 = float(item)
                            elif ljnn < 5:
                                cs12 = float(item)
                            elif ljnn < 6:
                                cs6 = float(item)
                                ljset = True
                            ljnn += 1
                    if ljset is True:
                        if self._ljparameters is None:
                            self._ljparameters = IndexedList()
                        self._ljparameters.append(
                                GromosNonbondedParam(
                                    type1=ilj-1, type2=jlj-1,
                                    type1_name=self._atomtypes[ilj-1], 
                                    type2_name=self._atomtypes[jlj-1],
                                    lj_r12=c12, lj_r6=c6,
                                    lj14_r12=cs12, lj14_r6=cs6)
                                )
                        ljnn=0
                        ilj = None
                        jlj = None
                        c12 = None
                        c6 = None
                        cs12 = None
                        cs6 = None
                        ljset = False
                    ljok=True
                if ljok:
                    ljn += 1
            if section == 'SOLUTEMOLECULES':
                smok=False
                if smn == 0:
                    smok=True
                    try:
                        val = cmdLine.split()[0]
                        smnum = int(val)
                        smnumset = True
                    except(TypeError, ValueError):
                        smnumset = False
                else:
                    if self.solmolecules is None:
                        self.solmolecules = []
                    for item in cmdLine.split():
                        self.solmolecules.append(int(item))
                    smok=True
                if smok:
                    smn += 1
            if section == 'TEMPERATUREGROUPS':
                tgok=False
                if tgn == 0:
                    tgok=True
                    try:
                        val = cmdLine.split()[0]
                        tgnum = int(val)
                        tgnumset = True
                    except(TypeError, ValueError):
                        tgnumset = False
                else:
                    if self.temperature_groups is None:
                        self.temperature_groups = []
                    for item in cmdLine.split():
                        self.temperature_groups.append(int(item))
                    tgok=True
                if tgok:
                    tgn += 1
            if section == 'PRESSUREGROUPS':
                pgok=False
                if pgn == 0:
                    pgok=True
                    try:
                        val = cmdLine.split()[0]
                        pgnum = int(val)
                        pgnumset = True
                    except(TypeError, ValueError):
                        pgnumset = False
                else:
                    if self.pressure_groups is None:
                        self.pressure_groups = []
                    for item in cmdLine.split():
                        self.pressure_groups.append(int(item))
                    pgok=True
                if pgok:
                    pgn += 1
            if section == 'LJEXCEPTIONS':
                ljeok=False
                if ljen == 0:
                    ljeok=True
                    try:
                        val = cmdLine.split()[0]
                        ljenum = int(val)
                        ljenumset = True
                    except(TypeError, ValueError):
                        ljenumset = False
                else:
                    if ljeset is False:
                        for item in cmdLine.split():
                            if ljenn < 1:
                                ilj = int(item)
                            elif ljenn < 2:
                                jlj = int(item)
                            elif ljenn < 3:
                                c12 = float(item)
                            elif ljnn < 4:
                                c6 = float(item)
                                ljeset = True
                            ljenn += 1
                    if ljeset is True:
                        if self._ljparameters is None:
                            self._ljparameters = IndexedList()
                            self._ljparameters.append(
                                    GromosNonbondedParam(
                                        type1=ilj-1, type2=jlj-1,
                                        type1_name=self._atomtypes[ilj-1], 
                                        type2_name=self._atomtypes[jlj-1],
                                        lj_r12=c12, lj_r6=c6,
                                        lj14_r12=cs12, lj14_r6=cs6)
                                    )
                        else:
                            nbfound=False
                            for ni, nb in enumerate(self._ljparameters):
                                if nb.type1 == ilj and nb.type2 == jlj:
                                    nbfound=True
                                    self._ljparameters[ni].lj_r12 = c12
                                    self._ljparameters[ni].lj_r6 = c6
                            if nbfound is False:
                                self._ljparameters.append(
                                        GromosNonbondedParam(
                                            type1=ilj-1, type2=jlj-1,
                                            type1_name=self._atomtypes[ilj-1], 
                                            type2_name=self._atomtypes[jlj-1],
                                            lj_r12=c12, lj_r6=c6)
                                        )
                        ljenn=0
                        ilj = None
                        jlj = None
                        c12 = None
                        c6 = None
                        ljeset = False
                    ljeok=True
                if ljeok:
                    ljen += 1
            if section == 'SOLVENTATOM':
                solvok=False
                if solvn == 0:
                    solvok=True
                    try:
                        val = cmdLine.split()[0]
                        solvnum = int(val)
                        solvnumset = True
                    except(TypeError, ValueError):
                        solvnumset = False
                else:
                    if solvset is False:
                        for item in cmdLine.split():
                            if solvnn == 0:
                                atnm = int(item)
                            elif solvnn == 1:
                                panm = item
                            elif solvnn == 2:
                                iac = int(item)
                            elif solvnn == 3:
                                mass = float(item)
                            elif solvnn == 4:
                                cg = float(item)
                                solvset = True
                            solvnn += 1
                    if solvset is True:
                        self.solvent = True
                        # check if SOLV in RESIDUES
                        # if not add it.
                        resfound = False
                        for resi, res in enumerate(self._residues):
                            if 'SOLV' in res:
                                mres = resi
                                resfound = True
                        if resfound is False:
                            self._residues.append('SOLV')
                            mres = len(self._residues)-1
                        if self.atomtypes is None:
                            self.atomtypes = IndexedList()
                        self.atomtypes.append(GromosAtomType(name=panm, 
                            _type=self._atomtypes[iac-1], type_id=iac, 
                            atom_number=atnm, mass=mass, charge=cg, 
                            solvent_atom=True,
                            res_id=mres, residue=self._residues[mres]))
                        solvnn=0
                        atnm = None
                        panm = None
                        iac = None
                        mass = None
                        cg = None
                        solvset = False
                    solvok=True
                if solvok:
                    solvn += 1
            if section == 'SOLVENTCONSTR':
                scok=False
                if scn == 0:
                    scok=True
                    try:
                        val = cmdLine.split()[0]
                        scnum = int(val)
                        scnumset = True
                    except(TypeError, ValueError):
                        scnumset = False
                else:
                    if scset is False:
                        for item in cmdLine.split():
                            if scnn < 1:
                                ic = int(item)
                            elif scnn < 2:
                                jc = int(item)
                            elif scnn < 3:
                                icb = float(item)
                                scset = True
                            scnn += 1
                    if scset is True:
                        if self.solvent_constraints is None:
                            self.solvent_constraints = IndexedList()
                        if self._bondparams is None:
                            self._bondparams = IndexedList()
                        # Adding this as a special bond type
                        # with solvent_constraint = True so
                        # we can track it as a pair constraint.
                        self._bondparams.append(GromosBondParam(
                            eq_length=icb, solvent_constraint=True))
                        # We are sure that it is used as a type.
                        self._bondparams[-1].used=True
                        # While we add the type as a bond, we do not 
                        # add the constraint as a bond so it will not 
                        # be miss classified as a bond. 
                        # First, we need to find the correct atom-type 
                        # indices
                        ib = None
                        jb = None
                        for atmi, atmt in enumerate(self.atomtypes):
                            if atmt.solvent_atom is True:
                                if ic == atmt.atom_number:
                                    ib = atmi+1
                                if jc == atmt.atom_number:
                                    jb = atmi+1
                        if ib is not None and jb is not None:
                            self.solvent_constraints.append(GromosBondType(
                                atom1=self.atomtypes[ib-1], 
                                atom2=self.atomtypes[jb-1], 
                                bond_param=self._bondparams[-1]
                                ))
                        scnn=0
                        ib = None
                        jb = None
                        icb = None
                        scset = False
                    scok=True
                if scok:
                    scn += 1

    def read_cnf(self, fin, decode=False):
        emptyLine = re.compile(r"^\s*$")
        gbline=0
        for line in fin:
            if decode:
                line = line.decode('utf-8')
            if emptyLine.findall(line):
                continue
            cmdLine = ' '.join([x for x in line.strip().split() if x])
            cmdLineUp = cmdLine.upper()
            if cmdLineUp.startswith('#'):
                continue
            if cmdLineUp.startswith('END'):
                section = None
            elif cmdLineUp.startswith('TITLE'):
                section = 'TITLE'
                continue
            elif cmdLineUp.startswith('TIMESTEP'):
                section = 'TIMESTEP'
                continue
            elif cmdLineUp.startswith('POSITION'):
                section = 'POSITION'
                continue
            elif cmdLineUp.startswith('LATTICESHIFTS'):
                section = 'LATTICESHIFTS'
                continue
            elif cmdLineUp.startswith('VELOCITY'):
                section = 'VELOCITY'
                continue
            elif cmdLineUp.startswith('GENBOX'):
                section = 'GENBOX'
                continue
            if section == 'TITLE':
                if self.conf_title is None:
                    self.conf_title = ''
                self.conf_title = self.conf_title + ' ' + cmdLine
            if section == 'TIMESTEP':
                if self.step is None:
                    self.step = [int(cmdLine.split()[0])]
                else:
                    self.step.append(int(cmdLine.split()[0]))
                if self.time is None:
                    self.time = [float(cmdLine.split()[1])]
                else:
                    self.time.append(float(cmdLine.split()[1]))
            if section == 'POSITION':
                if self.atoms is None:
                    self.atoms = IndexedList()
                if self.positions is None:
                    self.positions = []
                resid = int(cmdLine.split()[0])
                resname = cmdLine.split()[1].strip()
                rresname = resname.upper().replace('-', '').replace('+', '')
                atyp = cmdLine.split()[2].strip()
                aatyp = atyp.upper().replace('-', '').replace('+', '')
                anum = int(cmdLine.split()[3])
                self.positions.append([float(x) for x in cmdLine.split()[4:]])
                if 'SOLV' == resname.upper():
                    if self.atomtypes:
                        at = [atmi for atmi, atmt in enumerate(
                            self.atomtypes) if(aatyp == atmt.name.upper(
                                ).replace('-', '').replace('+', '') and 
                                resname == atmt.residue)]
                        self.atoms.append(GromosAtom(name=atyp, 
                            atom_type=self.atomtypes[at[0]], 
                            atom_number=anum, 
                            solvent_mol=True,
                            residue=resname, mol_id=resid))
                    else:
                        self.atoms.append(GromosAtom(name=atyp, 
                            atom_type=None, 
                            atom_number=anum, 
                            solvent_mol=True,
                            residue=resname, mol_id=resid))
                else:
                    if self.atomtypes:
                        at = [atmi for atmi, atmt in enumerate(
                            self.atomtypes) if(aatyp == atmt.name.upper(
                                ).replace('-', '').replace('+', '') and 
                                rresname == atmt.residue.upper(
                                    ).replace('-', '').replace('+', '') and 
                                resid == atmt.res_id)]
                        self.atoms.append(GromosAtom(name=atyp, 
                            atom_type=self.atomtypes[at[0]], 
                            atom_number=anum, 
                            residue=resname, mol_id=resid))
                    else:
                        self.atoms.append(GromosAtom(name=atyp, 
                            atom_type=None, 
                            atom_number=anum, 
                            residue=resname, mol_id=resid))
            if section == 'VELOCITY':
                if self.velocities is None:
                    self.velocities = []
                self.velocities.append([float(x) for x in cmdLine.split()[4:]])
                if self.positions is None:
                    if self.atoms is None:
                        self.atoms = IndexedList()
                    resid = int(cmdLine.split()[0])
                    resname = cmdLine.split()[1]
                    rresname = resname.upper().replace('-', '').replace('+', '')
                    atyp = cmdLine.split()[2]
                    aatyp = atyp.upper().replace('-', '').replace('+', '')
                    anum = int(cmdLine.split()[3])
                    if 'SOLV' in resname:
                        if self.atomtypes:
                            at = [atmi for atmi, atmt in enumerate(
                                self.atomtypes) if(aatyp == atmt.name.upper(
                                    ).replace('-', '').replace('+', '') and 
                                    resname == atmt.residue)]
                            self.atoms.append(GromosAtom(name=atyp, 
                                atom_type=self.atomtypes[at[0]], 
                                atom_number=anum, 
                                solvent_mol=True,
                                residue=resname, mol_id=resid))
                        else:
                            self.atoms.append(GromosAtom(name=atyp, 
                                atom_type=None, 
                                atom_number=anum, 
                                solvent_mol=True,
                                residue=resname, mol_id=resid))
                    else:
                        if self.atomtypes:
                            at = [atmi for atmi, atmt in enumerate(
                                self.atomtypes) if(aatyp == atmt.name.upper(
                                    ).replace('-', '').replace('+', '') and 
                                    rresname == atmt.residue.upper(
                                        ).replace('-', '').replace('+', '') and 
                                    resid == atmt.res_id)]
                            self.atoms.append(GromosAtom(name=atyp, 
                                atom_type=self.atomtypes[at[0]], 
                                atom_number=anum, 
                                residue=resname, mol_id=resid))
                        else:
                            self.atoms.append(GromosAtom(name=atyp, 
                                atom_type=None, 
                                atom_number=anum, 
                                residue=resname, mol_id=resid))
            if section == 'GENBOX':
                if gbline == 0:
                    self.nbounds = int(cmdLine.split()[0])
                elif gbline == 1:
                    self.a = float(cmdLine.split()[0])
                    self.b = float(cmdLine.split()[1])
                    self.c = float(cmdLine.split()[2])
                elif gbline == 2:
                    self.alpha = float(cmdLine.split()[0])
                    self.beta  = float(cmdLine.split()[1])
                    self.gamma = float(cmdLine.split()[2])
                elif gbline == 3:
                    self.phi = float(cmdLine.split()[0])
                    self.theta  = float(cmdLine.split()[1])
                    self.psi = float(cmdLine.split()[2])
                elif gbline == 4:
                    self.X = float(cmdLine.split()[0])
                    self.Y  = float(cmdLine.split()[1])
                    self.Z = float(cmdLine.split()[2])
                    gbline=0
                gbline += 1

    def __repr__(self):
        repname = '<GromosTopology' 
        if self.atoms:
            repname = repname + ' %d Atoms' % (len(self.atoms))
        if self.atomtypes:
            repname = repname + ', %d Atom Types' % (len(self.atomtypes))
        if self._residues:
            repname = repname + ', %d Residues' % (len(self._residues))
        if self.solvent:
            repname = repname + ' with Solvent'
        return repname + '>'

class GromosTrajOpen(object):

    def __init__(self, filename, handle):
        self.handle = handle
        self.filename = filename
        self.ftype = None
        self.ztype = None
        self.zfile = None
        self.ftype, self.zfile, self.ztype = get_zipType(self.filename)

    def __enter__(self):
        if self.handle.finTraj is not None:
            return self
        else:
            if self.ztype is not None:
                if 'zip' in self.ztype:
                    self.handle.zinTraj = zipfile.ZipFile(self.filename)
                    self.handle.finTraj = zin.open(self.zfile, 'r')
                    self.handle.decode = True
                    return self
                elif 'gz' in self.ztype:
                    self.handle.finTraj = gzip.open(self.filename, 'rt')
                    return self
            else:
                self.handle.finTraj = open(self.filename, 'r')
                return self

    def __exit__(self, type, value, traceback):
        if self.handle.finEOF:
            if self.handle.finTraj is not None:
                self.handle.finTraj.close()
            if self.handle.zinTraj is not None:
                self.handle.zinTraj.close()

    def iread_traj(self):
        title = None
        nbounds = None
        time = None
        step = None
        trajDict = None
        positions = None
        velocities = None
        a = None # box dimensions
        b = None
        c = None
        alpha = None # box angles
        beta = None
        gamma = None
        phi = None # rotation angles
        theta= None
        psi = None
        X = None
        Y = None
        Z = None
        section = None
        emptyLine = re.compile(r"^\s*$")
        gbline=0
        rtnOK=[False, False, False]
        while True:
            line = self.handle.finTraj.readline()
            if line:
                pass
            else:
                self.handle.finEOF = True
                break
            if self.handle.decode:
                line = line.decode('utf-8')
            if emptyLine.findall(line):
                continue
            cmdLine = ' '.join([x for x in line.strip().split() if x])
            cmdLineUp = cmdLine.upper()
            if cmdLineUp.startswith('#'):
                continue
            if cmdLineUp.startswith('END'):
                if trajDict is None:
                    trajDict = {}
                if section == "TITLE":
                    trajDict.update({'TITLE' : title})
                if section == "TIMESTEP":
                    trajDict.update({'STEP' : step})
                    trajDict.update({'TIME' : time})
                    rtnOK[0] = True
                if section == "POSITION":
                    trajDict.update({'POSITIONS' : np.asarray(positions)})
                    rtnOK[1] = True
                if section == "VELOCITY":
                    trajDict.update({'VELOCITIES' : np.asarray(velocities)})
                    rtnOK[1] = True
                if section == "GENBOX":
                    trajDict.update({'PHI' : phi})
                    trajDict.update({'THETA' : theta})
                    trajDict.update({'PSI' : psi})
                    trajDict.update({'X' : X})
                    trajDict.update({'Y' : Y})
                    trajDict.update({'Z' : Z})
                    trajDict.update({
                        'UNITCELL' : [
                            a, b, c, 
                            alpha, beta, gamma
                            ]
                        })
                    gbline=0
                    rtnOK[2] = True
                section = None
            elif cmdLineUp.startswith('TITLE'):
                section = 'TITLE'
                continue
            elif cmdLineUp.startswith('TIMESTEP'):
                section = 'TIMESTEP'
                continue
            elif cmdLineUp.startswith('POSITION'):
                section = 'POSITION'
                continue
            elif cmdLineUp.startswith('VELOCITY'):
                section = 'VELOCITY'
                continue
            elif cmdLineUp.startswith('GENBOX'):
                section = 'GENBOX'
                continue
            if section == 'TITLE':
                if title is None:
                    title = ''
                title = title + ' ' + cmdLine
            elif section == 'TIMESTEP':
                if step is None:
                    step = [int(cmdLine.split()[0])]
                else:
                    step.append(int(cmdLine.split()[0]))
                if time is None:
                    time = [float(cmdLine.split()[1])]
                else:
                    time.append(float(cmdLine.split()[1]))
            elif section == 'POSITION':
                if positions is None:
                    positions = []
                positions.append([float(x) for x in cmdLine.split()])
            elif section == 'VELOCITY':
                if velocities is None:
                    velocities = []
                velocities.append([float(x) for x in cmdLine.split()])
            elif section == 'GENBOX':
                if gbline == 0:
                    nbounds = int(cmdLine.split()[0])
                elif gbline == 1:
                    a = float(cmdLine.split()[0])
                    b = float(cmdLine.split()[1])
                    c = float(cmdLine.split()[2])
                elif gbline == 2:
                    alpha = float(cmdLine.split()[0])
                    beta  = float(cmdLine.split()[1])
                    gamma = float(cmdLine.split()[2])
                elif gbline == 3:
                    phi = float(cmdLine.split()[0])
                    theta  = float(cmdLine.split()[1])
                    psi = float(cmdLine.split()[2])
                elif gbline == 4:
                    X = float(cmdLine.split()[0])
                    Y  = float(cmdLine.split()[1])
                    Z = float(cmdLine.split()[2])
                gbline += 1
            if section is None:
                if all(rtnOK) == True:
                    return trajDict

class GromosTrajectory(GromosTopoObject):

    def __init__(self, trc=None, trv=None,
                 trc_format=None, trv_format=None): 
        self.trcfile = trc
        self.trvfile = trv
        self.trctype = trc_format
        self.trvtype = trv_format
        self.step = None
        self.time = None
        self.positions = None
        self.velocities = None
        self.unitcell = None
        self.trajiter = None
        self.trajDict = None
        self.decode = False
        self.finEOF = None
        self.finTraj = None
        self.zinTraj = None
        self.load()

    def load(self, clear=True):
        self.readTRCOK=False
        self.readTRVOK=False
        self.zctype=None
        self.zcfile=None
        self.zvtype=None
        self.zvfile=None
        if self.trctype is None:
            self.trctype, self.zcfile, self.zctype = get_zipType(self.trcfile)
        if('gromostrc' in self.trctype or
           'trc' in self.trctype):
            typtrc, self.zcfile, self.zctype = get_zipType(self.trcfile)
            if self.trcfile is not None:
                self.readTRCOK=True
        #if self.trvtype is None:
        #    self.trvtype, self.zvfile, self.zvtype = get_zipType(self.trcfile)
        #if('gromostrc' in self.trctype or
        #   'trc' in self.trctype):
        #    self.typtrv, self.zvfile, self.zvtype = get_zipType(self.trcfile)
        #    if self.trvfile is not None:
        #        self.readTRVOK=True

    def iter_read(self):
        theDict = None
        if self.readTRCOK:
            with GromosTrajOpen(self.trcfile, self) as f:
                while True:
                    theDict = f.iread_traj()
                    if theDict is not None:
                        yield theDict 
                    else:
                        break
        return None

    def iread(self):
        """Returns an iterator that goes through the given trajectory file one
           configuration at a time.
        """
        iterator_object = iter(self.iter_read())
        try:
            self.trajiter = next(iterator_object)
            if self.trajiter is not None:
                if 'STEP' in self.trajiter:
                    self.step = self.trajiter['STEP'][0]
                if 'TIME' in self.trajiter:
                    self.time = self.trajiter['TIME'][0]
                if 'POSITIONS' in self.trajiter:
                    self.positions = self.trajiter['POSITIONS']
                if 'VELOCITIES' in self.trajiter:
                    self.velocities = self.trajiter['VELOCITIES']
                if 'UNITCELL' in self.trajiter:
                    self.unitcell = self.trajiter['UNITCELL']
            return self.trajiter
        except StopIteration:
            pass
        finally:
            del iterator_object

if __name__ == "__main__":
    conffile=None
    conftype=None
    topofile=sys.argv[1]
    topotype='gromostop'
    if len(sys.argv)>2:
        conffile=sys.argv[2]
        conftype='gromoscnf'
    if len(sys.argv)>3:
        trcfile=sys.argv[3]
        trctype='gromostrc'
    #gtop = GromosTopology(top=topofile, cnf=conffile, 
    #        top_format=topotype, cnf_format=conftype)
    gtop = GromosTopology()
    gtop.load_top(topofile, topotype)
    gtop.load_cnf(conffile, conftype)
    print(gtop)
    #print(gtop.atomtypes)
    #print(gtop._residues)
    #if gtop.atomtypes:
    #    for atype in gtop.atomtypes:
    #        print([atype.unique_name]) 
    #        #print([atype.atom_number, atype.res_id, 
    #        #       atype.name, atype.type_id, 
    #        #       atype.mass, atype.charge,
    #        #       atype.charge_group, atype.exclude_atoms,
    #        #       atype.lj14_atoms])
#
#    print(gtop.atomtypes[10].list)
#    #print(gtop._bondtypes)
#    print("BONDTYPES:",gtop.bondtypes)
#    #if gtop._bondtypes:
#    #    for btype in gtop._bondtypes:
#    #        if btype.atoms_list:
#    #            print("BONDTYPE:",btype._id,btype.atoms_list) 
#    print("--------------------------------------------")
#    print("BONDPARAMS:",gtop.bondparams)
#    print("--------------------------------------------")
#    print("ANGLEPARAMS:",gtop.angleparams)
#    print("--------------------------------------------")
#    print("ANGLETYPES:",gtop.angletypes)
#    print("--------------------------------------------")
#    print("DIHPARAMS:",gtop.dihedralparams)
#    print("--------------------------------------------")
#    print("IMPPARAMS:",gtop.improperparams)
#    print("--------------------------------------------")
#    print("DIHEDRALTYPES:",gtop.dihedraltypes)
#    print("--------------------------------------------")
#    print("IMPROPERTYPES:",gtop.impropertypes)
#    print("--------------------------------------------")
#    print(gtop.atomtypes)
#    print("LJPARAMS:",gtop._ljparameters)
#    print("--------------------------------------------")
#    print("SOLVCONST:",gtop.solvent_constraints)
#    print("--------------------------------------------")
#    print("LJPARAMS:",gtop.ljparameters)
#    print("--------------------------------------------")
#    for res, rlist in gtop.residues.items():
#        print(res, rlist)
#    print("--------------------------------------------")
#    print(gtop.atoms)
    print("--------------------------------------------")
    print(np.asarray(gtop.positions))
    print("--------------------------------------------")
    gtraj = GromosTrajectory(trc=trcfile, trc_format=trctype) 
    pos = None
    while True:
        pos = gtraj.iread()
        if pos is not None:
            print("-- STEP: %s --" % (gtraj.step))
            print(gtraj.positions)
            print("UNITCELL:", gtraj.unitcell)
            print("---------------")
        else:
            break
             
